"""Base class for all dose engines."""

import logging

import sys

if sys.version_info < (3, 10):
    import importlib_resources as resources  # Backport for older versions
else:
    from importlib import resources  # Standard from Python 3.9+

import warnings
import time
from typing import ClassVar, Union
from abc import ABC, abstractmethod

import SimpleITK as sitk
import numpy as np
from pydantic import ValidationError

from pyRadPlan.core import Grid
from pyRadPlan.core.np2sitk import linear_indices_to_grid_coordinates
from pyRadPlan.core.resample import resample_numpy_array
from pyRadPlan.ct import CT, resample_ct
from pyRadPlan.cst import StructureSet
from pyRadPlan.stf import SteeringInformation, validate_stf
from pyRadPlan.plan import Plan, validate_pln
from pyRadPlan.dij import Dij, validate_dij
from pyRadPlan.scenarios import create_scenario_model, ScenarioModel
from pyRadPlan.machines import load_machine_from_mat, validate_machine

logger = logging.getLogger(__name__)

# from dose.calcDoseInit import init_dose_calc


class DoseEngineBase(ABC):
    """
    Abstract Interface for all dose engines.

    All dose engines should inherit from this class.

    Parameters
    ----------
    pln : Plan
        The Plan object to assign properties from.

    Attributes
    ----------
    short_name : str
        The short name of the dose engine.
    name : str
        The name of the dose engine.
    possible_radiation_modes : list[str]
        The possible radiation modes for the dose engine.
    is_dose_engine : bool = True
        Helper class variable telling you that this is a dose engine
    select_voxels_in_scenarios : bool
        Whether to select voxels in scenarios.
    mult_scen : Union[str, ScenarioModel]
        The scenario model.
    bio_model : Union[str, dict]
        The biological model.
    dose_grid : Union[Grid,dict]
        The dose grid to use (struct with at least doseGrid.resolution.x/y/z set).
    """

    # Constant, Abstract properties
    short_name: ClassVar[str]
    name: ClassVar[str]
    possible_radiation_modes: ClassVar[list[str]] = NotImplemented
    is_dose_engine: ClassVar[bool] = True  # Helper variable

    mult_scen: Union[str, ScenarioModel]
    bio_model: Union[str, dict]
    dose_grid: Union[Grid, dict]

    select_voxels_in_scenarios: bool

    # Public properties
    def __init__(self, pln: Union[Plan, dict] = None):
        # Assign default parameters from Matrad_Config or manually
        self.mult_scen = "nomScen"
        self.select_voxels_in_scenarios = None
        self.voxel_sub_ix = None  # selection of where to calculate / store dose, empty by default
        self.dose_grid = None

        if pln is not None:
            self.assign_properties_from_pln(pln, True)

        self._ct_grid = None

        # Protected properties with public get access
        self._machine = None  # base data defined in machine file
        self._timers = None  # timers of dose calc
        self._num_of_columns_dij = None  # number of columns in the dij struct
        self._vox_world_coords = None  # ct voxel coordinates in world
        self._vox_world_coords_dose_grid = None  # dose grid voxel coordinates in world
        self._vct_grid = None  # voxel grid inside patient
        self._vdose_grid = None  # voxel dose grid
        self._vct_grid_scen_ix = None  # logical subindexes of scenarios in ct grid
        self._vdose_grid_scen_ix = None  # logical subindexes of scenarios in dose grid
        self._vct_grid_mask = None  # voxel grid inside patient as logical mask
        self._vdose_grid_mask = None  # voxel dose grid inside patient as logical mask
        self._robust_voxels_on_grid = None  # voxels to be computed in robustness scenarios

        # Protected properties with private get access
        self._last_progress_update = None
        self._calc_dose_direct = False

        # initializing super() class (here: ABC)
        super().__init__()

    # Public methods
    def assign_properties_from_pln(self, pln: Plan, warn_when_property_changed: bool = False):
        """
        Assign properties from a Plan object to the Dose Engine.

        This includes the Scenario Model and the Biological Model, and any
        other properties that
        can be stored in the prop_dose_calc dictionary within the Plan object.
        This function will check if a property exists for the dose engine and,
        if yes, set it.

        Parameters
        ----------
        pln : Plan
            The Plan object to assign properties from.
        warn_when_property_changed : bool
            Whether to warn when properties are changed.
        """

        pln = validate_pln(pln)

        # Set Scenario Model
        if hasattr(pln, "mult_scen"):
            self.mult_scen = pln.mult_scen

        # Assign Biologival Model
        if hasattr(pln, "bio_param"):
            self.bio_param = pln.bio_param  # TODO: No bio_param yet

        if not isinstance(warn_when_property_changed, bool):
            warn_when_property_changed = False

        # Overwrite default properties within the engine
        # with the ones given in the prop_dose_calc dict
        if pln.prop_dose_calc is not None:
            prop_dict = pln.prop_dose_calc
            if (
                "engine" in prop_dict
                and prop_dict["engine"]
                and prop_dict["engine"] != self.short_name
            ):
                raise ValueError(
                    f"Inconsistent dose engines given! pln asks for '{prop_dict['engine']}', "
                    f"but you are using '{self.short_name}'!"
                )
            prop_dict.pop("engine", None)
        else:
            prop_dict = {}

        fields = prop_dict.keys()

        # Set up warning message
        if warn_when_property_changed:
            warning_msg = "Property in Dose Engine overwritten from pln.propDoseCalc"
        else:
            warning_msg = None

        for field in fields:
            if not hasattr(self, field):
                warnings.warn('Property "{}" not found in Dose Engine!'.format(field))
            elif warn_when_property_changed and warning_msg:
                logger.warning(warning_msg + f": {field}")

            setattr(self, field, prop_dict[field])

    def calc_dose_forward(
        self, ct: CT, cst: StructureSet, stf: SteeringInformation, w: np.ndarray
    ) -> dict:
        """
        Perform a forward dose calculation.

        Compute a dose cube by directly applying a set of weights
        during dose calculation.

        Parameters
        ----------
        ct : CT
            The CT scan data.
        cst : StructureSet
            The structure set containing volumes of interest (VOIs).
        stf : SteeringInformation
            The steering information containing beam configurations.
        w : np.ndarray
            The weights to apply during dose calculation.

        Returns
        -------
        Dij
            The dose information

        Notes
        -----
        pyRadPlan handles the forward dose calculation by setting a switch in
        the dose engine (_calc_dose_direct) to True. This facilitates reusing
        of algorithms in both forward and influence matrix calculations.
        """
        time_start = time.time()

        self._calc_dose_direct = True

        stf = validate_stf(stf)

        if w is None:
            logger.info("No weights given. Using weights stored in stf.")
        else:
            logger.info("Using provided weights for forward dose calculation.")

            if w.size != stf.total_number_of_bixels:
                raise ValueError("Number of weights does not match the number of bixels in stf.")

            # Assign the weights to the stf
            bixel_num = 0
            for beam in stf.beams:
                for ray in beam.rays:
                    for bixel in ray.beamlets:
                        bixel.weight = w[bixel_num]
                        bixel_num += 1

        # Run dose calculation (direct flag is on)
        dij = self._calc_dose(ct, cst, stf)

        # Now do the forward weighting with w
        # This is done because the engine might store the individual fields
        result = dij.compute_result_ct_grid(np.ones(dij.total_num_of_bixels, dtype=np.float32))

        time_elapsed = time.time() - time_start
        logger.info("Forward dose calculation done in %.2f seconds.", time_elapsed)

        return result

    def calc_dose_influence(self, ct: CT, cst: StructureSet, stf: SteeringInformation) -> Dij:
        """
        Calculate the set of dose/quantity influence matrices.

        These are the matrices that map a fluence vector to a dose/quantity
        distribution.

        Parameters
        ----------
        ct : CT
            The CT scan data.
        cst : StructureSet
            The structure set containing volumes of interest (VOIs).
        stf : SteeringInformation
            The steering information containing beam configurations.

        Returns
        -------
        Dij
            The dose influence matrix collection

        Notes
        -----
        pyRadPlan handles the forward dose calculation by setting a switch in
        the dose engine (_calc_dose_direct) to True. This facilitates reusing
        of algorithms in both forward and influence matrix calculations.
        """

        time_start = time.time()
        self._calc_dose_direct = False
        dij = self._calc_dose(ct, cst, stf)
        time_elapsed = time.time() - time_start
        logger.info("Dose influence matrix calculation done in %.2f seconds.", time_elapsed)

        return dij

    def set_overlap_priorities(self, cst: StructureSet, ct_dim=None) -> StructureSet:
        """Set overlap priorities for the structures in the CST."""

        logger.info("Adjusting structures for overlap... ")
        t_start = time.time()

        num_of_ct_scenarios = np.unique([x.num_of_scenarios for x in cst.vois])

        # Sanity check
        if len(num_of_ct_scenarios) > 1:
            raise ValueError("Inconsistent number of scenarios in cst struct.")

        new_cst = cst.apply_overlap_priorities()

        t_end = time.time()

        logger.info("Done in %.2f seconds.", t_end - t_start)

        return new_cst

    def select_voxels_from_cst(self, cst, dose_grid, selection_mode):
        """
        Get mask of the voxels (on dose grid).

        Gets the mask from the cst structures specified by selection_mode.
        """
        self.set_overlap_priorities(self, cst, dose_grid["dimensions"])

        selected_cst_structs = []

        include_mask = [
            np.zeros(np.prod(dose_grid["dimensions"]), dtype=bool) for _ in range(len(cst[:, 4]))
        ]

        if selection_mode == "all":
            for ct_scen_idx in range(len(include_mask)):
                include_mask[ct_scen_idx][:] = True
        else:
            if isinstance(selection_mode, str):
                if selection_mode == "targetOnly":
                    selected_cst_structs = [i for i, x in enumerate(cst[:, 2]) if x == "TARGET"]
                elif selection_mode == "objectivesOnly":
                    for i in range(len(cst)):
                        if len(cst[i, 5]) > 0:
                            selected_cst_structs.append(i)
                elif selection_mode == "oarsOnly":
                    selected_cst_structs = [i for i, x in enumerate(cst[:, 2]) if x == "OAR"]
                elif selection_mode == "robustnessOnly":
                    for i in range(len(cst)):
                        for j in range(len(cst[i, 5])):
                            if (
                                "robustness" in cst[i, 5][j]
                                and cst[i, 5][j]["robustness"] != "none"
                            ):
                                selected_cst_structs.append(i)
                else:
                    raise ValueError(f"Unrecognized voxel selection mode: {selection_mode}")
            elif isinstance(selection_mode, (list, np.ndarray)):
                selected_cst_structs = np.unique(
                    np.intersect1d(selection_mode, np.arange(len(cst)) + 1)
                )
                if not np.array_equal(selected_cst_structs, np.unique(selection_mode)):
                    warnings.warn(
                        "Specified structures are not compatible with cst structures. "
                        f"Only performing calculation on structures: {selected_cst_structs}"
                    )

            # Loop over all cst structures
            for i in range(len(cst)):
                if len(cst[i, 4][0]) > 0:
                    if len(cst[i, 5]) > 0:
                        # Loop over obj/constraint functions
                        for j in range(len(cst[i, 5])):
                            obj = cst[i, 5][j]

                            if not isinstance(obj, dict):
                                try:
                                    pass
                                    # TODO: obj = matRad_DoseOptimizationFunction.
                                    # createInstanceFromStruct(obj)
                                except ValueError:
                                    logger.error(
                                        f"cst[{i},5][{j}] is not a valid Objective/constraint! "
                                        "Remove or Replace and try again!"
                                    )

                            robustness = obj.get("robustness", "none")

                            if i in selected_cst_structs:
                                for ct_idx in range(len(cst[i, 4])):
                                    include_mask[ct_idx][cst[i, 4][ct_idx]] = 1

                                if robustness == "none":
                                    logger.info(
                                        f"Including cst structure {cst[i, 1]} even though this "
                                        "structure has no robustness."
                                    )
                            else:
                                logger.info(
                                    f"Excluding cst structure {cst[i, 1]} even though this "
                                    "structure has an objective or constraint."
                                )

                                if robustness != "none":
                                    logger.info(
                                        f"Excluding cst structure {cst[i, 1]} even though this "
                                        "structure has robustness."
                                    )
                    elif i in selected_cst_structs:
                        logger.info(
                            f"Including cst structure {cst[i, 1]} even though this structure "
                            "does not have any objective or constraint"
                        )

        return include_mask

    def resize_cst_to_grid(self, cst: StructureSet, dij, new_ct: CT) -> StructureSet:
        """Resize the CST to the dose cube resolution."""

        logger.info("Resampling structure set... ")

        t_start = time.time()

        cst.resample_on_new_ct(new_ct)

        t_end = time.time()

        logger.info("Done in  %.2f seconds.", t_end - t_start)
        return cst

    # private methods
    def _init_dose_calc(self, ct: CT, cst: StructureSet, stf: SteeringInformation) -> dict:
        """
        Initialize the dose calculation.

        Parameters
        ----------
        self : DoseEngineBase
            The instance of the dose engine.
        ct : CT
            The CT scan data.
        cst : StructureSet
            The structure set containing volumes of interest (VOIs).
        stf : SteeringInformation
            The steering information containing beam configurations.

        Returns
        -------
        dict
            A dictionary containing the dose influence matrix (dij) and related information.

        Description
        -----------
        This method sets up the necessary grids and parameters for dose calculation. It initializes
        the CT and dose grids, checks machine and radiation mode consistency, creates a scenario
        model, and sets up arrays for bookkeeping. It also adjusts the isocenter internally for
        different dose grids and converts CT subscripts to world coordinates. Additionally, it
        loads the machine file from the base data folder and performs voxel selection for dose
        calculation.
        """

        self._ct_grid = Grid.from_sitk_image(ct.cube_hu)

        dij = {}

        # Default: dose influence matrix computation
        if self._calc_dose_direct:
            logger.info("Forward dose calculation using '%s' Dose Engine...", self.name)
        else:
            logger.info("Dose influence matrix calculation using '%s' Dose Engine...", self.name)

        # Check if machine and radiation_mode are consistent
        machine = list({beam.machine for beam in stf.beams})
        radiation_mode = list({beam.radiation_mode for beam in stf.beams})

        if len(machine) != 1 or len(radiation_mode) != 1:
            raise ValueError("machine and radiation mode need to be unique within supplied stf!")

        machine = machine[0]
        radiation_mode = radiation_mode[0]

        # ScenarioModel
        if not isinstance(self.mult_scen, ScenarioModel):
            self.mult_scen = create_scenario_model(self.mult_scen, ct)

        # TODO: Add BioModel Support
        # if not np.isnan(self.bio_param["RBE"]):
        #     dij['RBE'] = self.bio_param['RBE']

        # store CT grid
        dij["ct_grid"] = ct.grid

        if self.dose_grid is None:
            logger.info("Dose Grid not set. Using default resolution.")
            self.dose_grid = ct.grid.resample({"x": 5.0, "y": 5.0, "z": 5.0})

        if not isinstance(self.dose_grid, Grid):
            try:
                self.dose_grid = Grid.model_validate(self.dose_grid)
            except ValidationError:
                self.dose_grid = ct.grid.resample(self.dose_grid["resolution"])

        logger.info(
            "Dose Grid has Dimensions (%d,%d,%d) with resolution x=%f, y=%f, z=%f.",
            self.dose_grid.dimensions[0],
            self.dose_grid.dimensions[1],
            self.dose_grid.dimensions[2],
            self.dose_grid.resolution["x"],
            self.dose_grid.resolution["y"],
            self.dose_grid.resolution["z"],
        )

        dij["dose_grid"] = self.dose_grid

        # Meta information for dij #TODO: Change dij Model in future
        dij["num_of_beams"] = len(stf.beams)
        dij["num_of_scenarios"] = 1  # TODO: Add support for multiple scenarios
        dij["num_of_rays_per_beam"] = [stf.beams[i].num_of_rays for i in range(len(stf.beams))]
        dij["total_num_of_bixels"] = int(
            sum([stf.beams[i].total_number_of_bixels for i in range(len(stf.beams))])
        )
        dij["total_num_of_rays"] = int(sum(dij["num_of_rays_per_beam"]))

        # Check if full dose influence data is required
        self._num_of_columns_dij = (
            len(stf.beams) if self._calc_dose_direct else dij["total_num_of_bixels"]
        )

        # Set up arrays for book keeping
        dij["bixel_num"] = np.nan * np.ones(self._num_of_columns_dij)
        dij["ray_num"] = np.nan * np.ones(self._num_of_columns_dij)
        dij["beam_num"] = np.nan * np.ones(self._num_of_columns_dij)

        # Default MU calibration
        dij["min_mu"] = np.zeros(self._num_of_columns_dij)
        dij["max_mu"] = np.ones(self._num_of_columns_dij) * np.inf
        dij["num_of_particles_per_mu"] = np.ones(self._num_of_columns_dij) * 1e6

        if self.voxel_sub_ix is None:
            self.voxel_sub_ix = np.unique(
                np.concatenate([cst.vois[c].indices_numpy for c in range(0, len(cst.vois))])
            )

        tmp_vct_grid_scen = [np.asarray(self.voxel_sub_ix)] * ct.num_of_ct_scen

        self._vct_grid = np.unique(np.concatenate(tmp_vct_grid_scen))

        # Find subindexes for the individual scenarios. This helps with subselection later.
        self._vct_grid_scen_ix = [np.isin(self._vct_grid, c) for c in tmp_vct_grid_scen]

        # Initialize tmp_vdose_grid_scen
        tmp_vdose_grid_scen = [None] * ct.num_of_ct_scen

        if self._ct_grid == self.dose_grid:
            tmp_vdose_grid_scen = tmp_vct_grid_scen
            resampled_ct = ct
        else:
            resampled_ct = resample_ct(
                ct=ct,
                interpolator=sitk.sitkNearestNeighbor,
                target_grid=self.dose_grid,
            )
            self.dose_grid = Grid.from_sitk_image(resampled_ct.cube_hu)

            for s in range(ct.num_of_ct_scen):
                # Receive linear indices and grid locations from the dose grid
                tmp_cube = np.zeros(ct.cube_dim[::-1], dtype=np.float32)
                tmp_cube.ravel()[tmp_vct_grid_scen[s]] = 1

                interpolated_tmp_cube = resample_numpy_array(
                    tmp_cube,
                    reference_image=ct.cube_hu,
                    interpolator=sitk.sitkNearestNeighbor,
                    target_image=resampled_ct.cube_hu,
                )

                tmp_vdose_grid_scen[s] = np.where(
                    np.abs(np.round(interpolated_tmp_cube)).ravel().astype(bool)
                )

        # Get unique dose grid
        self._vdose_grid = np.unique(np.concatenate(tmp_vdose_grid_scen, axis=0))

        # Find subindexes for the dose grid scenarios
        self._vdose_grid_scen_ix = [np.isin(self._vdose_grid, c) for c in tmp_vdose_grid_scen]

        # Convert CT subscripts to world coordinates.
        self._vox_world_coords = linear_indices_to_grid_coordinates(
            self._vct_grid, self._ct_grid, index_type="numpy"
        )

        self._vox_world_coords_dose_grid = linear_indices_to_grid_coordinates(
            self._vdose_grid, self.dose_grid, index_type="numpy"
        )

        # Create helper masks
        self._vdose_grid_mask = np.zeros(self.dose_grid.num_voxels, dtype=bool)
        self._vdose_grid_mask[self._vdose_grid] = True

        self._vct_grid_mask = np.zeros(np.prod(ct.cube_dim), dtype=bool)
        self._vct_grid_mask[self._vct_grid] = True

        # Load machine file from base data folder
        self._machine = self.load_machine(radiation_mode, machine)

        # Voxel selection for dose calculation
        # TODO: set overlap priorites - Skips most of the function
        cst = self.set_overlap_priorities(cst, ct_dim=None)

        # resizing cst to dose cube resolution
        cst = self.resize_cst_to_grid(cst, dij, resampled_ct)

        # TODO: this function has not yet been implemented
        # structures that are selected here will be included in dose calculation over
        # the robust scenarios
        # self._robust_voxels_on_grid = select_voxels_from_cst(cst, dij["ct_grid"])

        return dij

    def _finalize_dose(self, dij: dict) -> Dij:
        return validate_dij(dij)

    def _progress_update(self, pos, total):
        raise NotImplementedError

    # Private and abstract methods

    @abstractmethod
    def _calc_dose(self, ct: CT, cst: StructureSet, stf: SteeringInformation) -> Dij:
        raise NotImplementedError("Method '_calc_dose' must be implemented.")

    # static methods
    @staticmethod
    def is_available(pln, machine):
        available, msg = [], []
        return available, msg

    @staticmethod
    def load_machine(radiation_mode, machine_name):
        file_name = radiation_mode + "_" + machine_name + ".mat"
        machines_path = resources.files("pyRadPlan.data.machines")
        path = machines_path.joinpath(file_name)
        machine = validate_machine(load_machine_from_mat(path))
        return machine



import numpy as np


def default_hlut(radiation_mode: str = None) -> np.ndarray:
    """Return default HLUT by radiation mode.

    Parameters
    ----------
    radiation_mode : str
        The radiation mode.

    Returns
    -------
    np.ndarray
        The default HLUT.

    Notes
    -----
    Currently, we provide a single hlut independent of the radiation mode.
    """

    hus = [
        [-1024, -999, -90, -45, 0, 100, 350, 3000],
        [0.001, 0.001, 0.95, 0.99, 1, 1.095, 1.199, 2.505],
    ]

    return np.array(hus).transpose()



from typing import Any
import numpy as np
from pydantic import Field, field_serializer, SerializerFunctionWrapHandler, FieldSerializationInfo
from pyRadPlan.core import PyRadPlanBaseModel


class RangeShifter(PyRadPlanBaseModel):
    """
    A class representing a range shifter.

    This class extends Pydantic's BaseModel and provides functionality to
    handle the range shifter information, including properties like id, equivalent thickness,
    and source to range shifter distance.

    Attributes
    ----------
    id : int
        The identifier for the range shifter.
    eq_thickness : float
        The equivalent thickness of the range shifter.
    source_rashi_distance : float
        The distance from the source to the range shifter.
    """

    id: int = Field(alias="ID", default=0)
    eq_thickness: float = Field(default=0.0)
    source_rashi_distance: float = Field(default=0.0)

    @field_serializer("*", mode="wrap")
    def _field_typing(
        self, v: Any, handler: SerializerFunctionWrapHandler, info: FieldSerializationInfo
    ) -> Any:
        """Serialize with possible matRad context."""

        context = info.context
        if context and context.get("matRad") == "mat-file":
            return np.float64(v)  # Ensure double for MATLAB/matRad

        return handler(v, info)



from typing import Union, cast
from pyRadPlan.ct import CT, validate_ct
from pyRadPlan.cst import StructureSet, validate_cst
from pyRadPlan.plan import validate_pln
from pyRadPlan.stf import (
    SteeringInformation,
    validate_stf,
)
from pyRadPlan.stf.generators import StfGeneratorBase
from pyRadPlan.stf import get_generator


def generate_stf(
    ct: Union[CT, dict], cst: Union[StructureSet, dict], pln: Union[StructureSet, dict]
) -> SteeringInformation:
    ct = validate_ct(ct)
    cst = validate_cst(cst, ct=ct)
    pln = validate_pln(pln)

    stfgen = get_generator(pln)

    stfgen = cast(StfGeneratorBase, stfgen)

    return validate_stf(stfgen.generate(ct, cst))



import warnings
import logging
from typing import Union, Type
from pyRadPlan.plan import validate_pln, Plan

from ._base import StfGeneratorBase


STF_GENERATORS = {}

logger = logging.getLogger(__name__)


def register_generator(generator_cls: Type[StfGeneratorBase]) -> None:
    """
    Register a new stf generator for irradiation geometry.

    Parameters
    ----------
    generator_cls : type
        A Generator class.
    """
    if not issubclass(generator_cls, StfGeneratorBase):
        raise ValueError("Generator must be a subclass of StfGeneratorBase.")

    if generator_cls.short_name is None:
        raise ValueError("Generator must have a 'short_name' attribute.")

    if generator_cls.name is None:
        raise ValueError("Generator must have a 'name' attribute.")

    if generator_cls.possible_radiation_modes is None:
        raise ValueError("Generator must have a 'possible_radiation_modes' attribute.")

    generator_name = generator_cls.short_name
    if generator_name in STF_GENERATORS:
        warnings.warn(f"Generator '{generator_name}' is already registered.")
    else:
        STF_GENERATORS[generator_name] = generator_cls


def get_available_generators(pln: Union[Plan, dict[str]]) -> dict[str, Type[StfGeneratorBase]]:
    """
    Get a list of available stf generators based on the plan.

    Parameters
    ----------
    pln : Union[Plan,dict]
        A Plan object.

    Returns
    -------
    list
        A list of available generators.
    """
    pln = validate_pln(pln)
    return {
        name: cls
        for name, cls in STF_GENERATORS.items()
        if pln.radiation_mode in cls.possible_radiation_modes
    }


def get_generator(pln: Union[Plan, dict]) -> StfGeneratorBase:
    """
    Get the appropriate generator based on the plan.

    Parameters
    ----------
    pln : Plan
        A Plan object.

    Returns
    -------
    Generator
        A Dose Generator object.
    """
    pln = validate_pln(pln)

    # Available generators
    generators = get_available_generators(pln)
    if len(generators) <= 0:
        raise ValueError(f"No generator available for radiation mode '{pln.radiation_mode}'.")

    generator_names = list(generators.keys())

    # Did the user provide an generator in the pln?
    if isinstance(pln.prop_stf, StfGeneratorBase):
        # The user provided an generator object, so lets use it
        # but warn the user if it is not in the available generators
        generator_name = pln.prop_stf.short_name
        if generator_name not in generators:
            warnings.warn(f"Generator '{generator_name}' seems not to be valid for Plan setup.")
        return pln.prop_stf

    if isinstance(pln.prop_stf, dict):
        # The user provided a dictionary with generator parameters, so we need to find the generator name
        if "generator" in pln.prop_stf:
            if pln.prop_stf["generator"] in generators:
                return generators[pln.prop_stf["generator"]](pln)
            warnings.warn(f"Generator '{pln.prop_stf['generator']}' not available for Plan.")

        # If no generator name was found, we choose the first as default
        logger.warning(
            "No generator specified in Plan. Using first available generator %s.",
            generator_names[0],
        )
        return generators[generator_names[0]](pln)

    raise ValueError(f"No generator available for radiation mode '{pln.radiation_mode}'.")



"""Handlers for Matlab files."""

import os
from pyRadPlan.io import matfile


class MatlabFileHandler:
    """Handler for Matlab files."""

    def __init__(self, temp_path):
        self.tempPath = temp_path

    def load(self, *args):
        mat_file_contents = {}
        for arg in args:
            mat_file_contents[arg] = matfile.load(os.path.join(self.tempPath, f"{arg}.mat"))
        return mat_file_contents

    def save(self, **kwargs):
        for key, value in kwargs.items():
            if not isinstance(value, dict):
                tmp_value = {key: value}
            else:
                tmp_value = value
            matfile.save(os.path.join(self.tempPath, f"{key}.mat"), tmp_value)

    def delete(self, *args):
        for arg in args:
            file_path = os.path.join(self.tempPath, f"{arg}.mat")
            os.remove(file_path)
            setattr(self, arg, {})



"""Provides basic classes for generating STF for ion beams."""

import warnings
import numpy as np

from pyRadPlan.raytracer import RayTracerSiddon
from pyRadPlan.machines import IonAccelerator
from pyRadPlan.ct import default_hlut

from .._rangeshifter import RangeShifter
from .._beamlet import IonSpot
from .._ray import Ray
from ._externalbeam import StfGeneratorExternalBeamRayBixel


class StfGeneratorIonRayBixel(StfGeneratorExternalBeamRayBixel):
    """
    Intermediate Interface for an Ion-Ray-Bixel Geometry Generator.

    Attributes
    ----------
    use_range_shifter : bool
        Use range shifter when finding energies for irradiation.
    use_given_wet_image : bool
        Use a provided WET image for range determination.
    """

    use_range_shifter: bool
    use_given_wet_image: bool

    def __init__(self, pln=None):
        self.use_range_shifter = False
        self.use_given_wet_image = False
        self._wet_image = None
        super().__init__(pln)

    def _initialize(self):
        super()._initialize()

        if not isinstance(self.machine, IonAccelerator):
            raise ValueError("Machine must be an instance of IonAccelerator")

        self._available_peak_positions = self.machine.peak_positions
        self._available_energies = self.machine.energies

        if self.use_range_shifter:
            range_shifter_wepl = np.round(self._available_peak_positions.min() * 1.25)
            self._available_peak_positions_range_shifter = (
                self.machine.peak_positions - range_shifter_wepl
            )
            warnings.warn(
                "Use of range shifter was enabled."
                "For now, pyRadPlan will generate generic range shifter "
                f"with WEPL {range_shifter_wepl} mm to enable ranges below "
                "the minimum energy in the machine."
            )

        if self.use_given_wet_image:
            if self._ct.cube is None:
                warnings.warn("No WET CT provided provided in CT. Cannot use given WET image.")
                self.use_given_wet_image = False

        if self.use_given_wet_image:
            self._wet_image = self._ct.cube
        else:
            # TODO: load HLUT for given scanner
            self._wet_image = self._ct.compute_wet(default_hlut(self.radiation_mode))


class StfGeneratorIonSingleSpot(StfGeneratorIonRayBixel):
    """Class representing a Ion Single Spot Geometry Stf Generator.

    Attributes
    ----------
    name : str
        The name of the generator ("Ion Single Spot").
    short_name : str
        The short name of the generator ("ionSingleSpot").
    possible_radiation_modes : list[str]
        A list of possible radiation modes.
    """

    name = "Ion Single Spot"
    short_name = "ionSingleSpot"
    possible_radiation_modes = ["protons", "helium", "carbon", "oxygen"]

    def _generate_ray_positions_in_isocenter_plane(self, beam):
        """Generate the ray positions in the isocenter plane.

        As we have a single spot, this is a single ray to the isocenter.

        Parameters
        ----------
        beam : dict
            The beam dictionary. Ignored in this case

        Returns
        -------
        np.ndarray
            A numpy array with the single ray position of (0, 0, 0) in the
            isocenter.
        """

        return np.zeros((3, 1), dtype=float)

    def _generate_source_geometry(self):
        """Generate the source geometry for the ion single spot geometry."""
        stf = super()._generate_source_geometry()

        return stf


class StfGeneratorIMPT(StfGeneratorIonRayBixel):
    """Class representing an Ion IMPT Geometry Stf Generator.

    Attributes
    ----------
    name : str
        The name of the generator ("Photon IMRT Geometry").
    short_name : str
        The short name of the generator ("photonIMRT").
    possible_radiation_modes : list[str]
        A list of possible radiation modes (["photons"]).
    """

    name = "IMPT"
    short_name = "IMPT"
    possible_radiation_modes = ["protons", "helium", "carbon", "oxygen"]

    # Alias for bixel_width
    @property
    def lateral_spot_spacing(self) -> float:
        """Alias for the bixel_width property."""
        return self.bixel_width

    @lateral_spot_spacing.setter
    def lateral_spot_spacing(self, value: float):
        self.bixel_width = value

    def __init__(self, pln=None):
        self.radiation_mode = "protons"
        self.longitudinal_spot_spacing = 3.0  # Longitudinal spot spacing in mm
        super().__init__(pln)

    def _generate_source_geometry(self):
        """Generate the source geometry for the photon IMRT geometry."""

        # TODO: Get available Energies
        stf = super()._generate_source_geometry()
        return stf

    def _create_rays(self, beam: dict) -> list[dict]:
        rays = super()._create_rays(beam)

        # Perform vectorized Ray Tracing from source through ray_pos through whole image
        trace_cubes = [self._wet_image, self._target_mask]
        rt = RayTracerSiddon(trace_cubes)

        ray_pos_all = np.zeros((len(rays), 3))
        for i, ray in enumerate(rays):
            ray_pos_all[i, :] = ray["ray_pos"]

        ray_pos_all_bev = np.zeros((len(rays), 3))
        for i, ray in enumerate(rays):
            ray_pos_all_bev[i, :] = ray["ray_pos_bev"]

        target_points = 2 * (ray_pos_all - beam["source_point"]) + beam["source_point"]
        target_points_bev = (
            2 * (ray_pos_all_bev - beam["source_point_bev"]) + beam["source_point_bev"]
        )
        # TODO: do we need to check scenarios here? Compare to matrad...
        _, lengths, rhos, _, _ = rt.trace_rays(
            beam["iso_center"],
            beam["source_point"].reshape((1, 3)),
            target_points,
        )

        for r, ray in enumerate(rays):
            # check if ray hit target

            # valid_alphas = np.isnan(alphas[r, :]) == False
            valid_ixs = np.isfinite(rhos[0][r, :])
            # alpha = alphas[r, valid_alphas]
            length = lengths[r, valid_ixs]
            rho = [cube_rho[r, valid_ixs] for cube_rho in rhos]
            # d12 = d12s[r]
            # ix = ixs[valid_values]

            ray_hit_target = np.count_nonzero(rho[1]) > 0

            ray_energies = np.array([])

            if ray_hit_target:
                ray["target_point"] = target_points[r]
                ray["target_point_bev"] = target_points_bev[r]
                # ct_entry_point = alpha[0] * d12

                # Obtain radiological depths / WET
                # rad_depths = np.cumsum(l * rho[0])
                rad_depths = np.cumsum(length * rho[0]) - length * rho[0] / 2.0

                rho_enter = np.insert(rho[1], 0, 0)
                # Use diff to find next voxel with different value
                diff_voi_enter = np.diff(rho_enter)
                entry_ix = np.argwhere(diff_voi_enter == 1)
                # Manage Edge case for last voxel
                entry_ix[entry_ix > len(rho[1])] = len(rho[1]) - 1
                # target entry at beginning of voxel
                target_entry = rad_depths[entry_ix] - length[entry_ix] * rho[0][entry_ix] / 2.0
                target_entry = target_entry.squeeze(axis=1)

                rho_exit = np.append(rho[1], 0)
                diff_voi_exit = np.diff(rho_exit)
                exit_ix = np.argwhere(diff_voi_exit == -1)
                # Manage Edge case for last voxel
                exit_ix[exit_ix > len(rho[1])] = len(rho[1]) - 1
                target_exit = rad_depths[exit_ix] + length[entry_ix] * rho[0][entry_ix] / 2.0
                target_exit = target_exit.squeeze(axis=1)

                # Sanity Check
                if target_entry.size != target_exit.size:
                    raise ValueError(
                        "Did not find same number of target entries and exits! "
                        "Can not assign energies for IMPT geometry."
                    )

                for i in range(target_entry.size):
                    # TODO: use range shifter

                    # Select energies
                    select_energies = self._available_energies[
                        np.logical_and(
                            self._available_peak_positions >= target_entry[i],
                            self._available_peak_positions <= target_exit[i],
                        )
                    ]

                    ray_energies = np.append(ray_energies, select_energies)

                # TODO: finalize range shifters
                ray_range_shifters = [RangeShifter() for _ in ray_energies]

                # TODO: select foci from available ones using the LUT
                ray_foci = np.zeros_like(ray_energies, dtype=int)

                # TODO: finalize MU settings
                ray_num_particles_per_mu = np.ones_like(ray_energies) * 1.0e6
                ray_min_mu = np.zeros_like(ray_energies)
                ray_max_mu = np.ones_like(ray_energies) * float("inf")
            else:
                ray_energies = np.array([])
                ray_foci = np.array([])

            beamlets = []
            # Create / Validate Beamlets
            for i, energy in enumerate(ray_energies):
                beamlet = IonSpot(
                    energy=energy,
                    num_particles_per_mu=ray_num_particles_per_mu[i],
                    min_mu=ray_min_mu[i],
                    max_mu=ray_max_mu[i],
                    range_shifter=ray_range_shifters[i],
                    focus_ix=ray_foci[i],
                )
                beamlets.append(beamlet)

            ray["beamlets"] = beamlets

        # Clean out empty rays with no energy
        rays = [ray for ray in rays if len(ray["beamlets"]) > 0]

        # Validate rays
        for r, ray in enumerate(rays):
            rays[r] = Ray.model_validate(ray)

        return rays



from typing import Any, Union, ClassVar
import warnings
import numpy as np

import SimpleITK as sitk

from pyRadPlan.core import np2sitk
from pyRadPlan.stf._exceptions import GeometryError
from pyRadPlan.plan import validate_pln
from pyRadPlan.ct import validate_ct, CT
from pyRadPlan.cst import validate_cst, StructureSet
from pyRadPlan.scenarios import ScenarioModel, validate_scenario_model
from pyRadPlan.machines import Machine, load_from_name, validate_machine


from abc import ABC, abstractmethod


class StfGeneratorBase(ABC):
    """Base class for steering information generators."""

    # Class constants
    is_stf_generator: ClassVar[bool] = True
    name: ClassVar[str]
    short_name: ClassVar[str]
    possible_radiation_modes: ClassVar[list[str]]

    @property
    def radiation_mode(self):
        """Radiation Mode."""
        return self._radiation_mode

    @radiation_mode.setter
    def radiation_mode(self, value):
        if value in self.possible_radiation_modes:
            self._radiation_mode = value
        else:
            raise ValueError(
                f"Invalid radiation mode. Possible modes are: {self.possible_radiation_modes}"
            )

    def _computed_target_margin(self) -> float:
        """Margin to be applied to the union of targets for stf generation."""
        return 0.0

    # Initialization method for non-constant properties
    def __init__(self, pln=None):
        # Defaults
        self.vis_mode: int = 0  # Visualization Options
        self.add_margin: bool = (
            True  # Add margins to target (for numerical stability and robustness)
        )

        self.mult_scen: Any = None  # Scenario Model, #TODO: Merge Scenario Model Developments
        self.bio_param: Any = None  # Biological Model #TODO: Merge Biological Model Developments
        self.machine: Union[str, Machine] = None  # Machine name
        self._ct: CT = None  # CT Object
        self._cst: StructureSet = None  # CST Object, TODO: Merge CST development
        self._mult_scen: ScenarioModel = None  # Scenario Model Object validated from mult_scen
        self._target_voxels: np.ndarray = None  # Target Voxels
        self._patient_voxels: np.ndarray = None  # Patient Voxels
        self._target_voxel_coordinates: np.ndarray = None  # Target Voxel Coordinates
        self._patient_voxel_coordinates: np.ndarray = None  # Patient Voxel Coordinates
        self._target_mask: sitk.Image = None  # Target Mask
        self._patient_mask: sitk.Image = None  # Patient Mask
        self._cube_hu: sitk.Image = None  # HU Cube pre-processed for stf generation

        # Validate plan to make sure it is a Plan instance
        if pln is not None:
            pln = validate_pln(pln)

            self._assign_parameters_from_pln(pln)

    def generate(self, ct, cst):
        """Generate the STF for the given CT and CST."""

        self._ct = validate_ct(ct)
        self._cst = validate_cst(cst)

        self._initialize()

        self._initialize_patient_geometry()
        stf = self._generate_source_geometry()

        # TODO: stf = validate_stf(stf)
        return stf

    def _initialize(self):
        """Initializion Procedure."""

        # Validate the Machine
        if isinstance(self.machine, str):
            tmp_machine = load_from_name(self.radiation_mode, self.machine)
        else:
            tmp_machine = self.machine

        tmp_machine = validate_machine(tmp_machine)

        if tmp_machine.radiation_mode != self.radiation_mode:
            raise ValueError(
                f"Machine radiation mode {tmp_machine.radiation_mode} does not match the stf "
                f"generators radiation mode {self.radiation_mode}"
            )

        self.machine = tmp_machine

        # Validate the scenario model
        self._mult_scen = validate_scenario_model(self.mult_scen, self._ct)

    def _initialize_patient_geometry(self):
        """Initialize patient geometry."""

        # Get all Patient Voxels
        self._patient_voxels = self._cst.patient_voxels(order="numpy")

        if not self._patient_voxels.any():
            raise GeometryError("Contours do not contain any voxels")

        self._patient_mask = self._cst.patient_mask()

        # now process target

        self._target_mask = self._cst.target_union_mask()
        # Add margin if desired, otherwise take the indices directly
        if not self.add_margin:
            self._target_voxels = self._cst.target_union_voxels(order="numpy")
        if self.add_margin:
            added_margin = self._computed_target_margin()  # pbMargin in matRad

            max_iso_shift = np.max(self._mult_scen.iso_shift, axis=0)
            range_margin = (
                self._mult_scen.max_abs_range_shift + self._mult_scen.max_rel_range_shift
            )

            constant_margin = added_margin + range_margin

            # Dimension specific margins
            dim_margins = (
                np.max((self._ct.resolution["x"], max_iso_shift[0], constant_margin)),
                np.max((self._ct.resolution["y"], max_iso_shift[1], constant_margin)),
                np.max((self._ct.resolution["z"], max_iso_shift[2], constant_margin)),
            )

            voxel_margin = np.ceil(
                np.array(dim_margins)
                / (self._ct.resolution["x"], self._ct.resolution["y"], self._ct.resolution["z"])
            ).astype(int)

            dilation = sitk.BinaryDilateImageFilter()
            dilation.SetKernelType(sitk.sitkBox)
            dilation.SetKernelRadius(voxel_margin.astype(int).tolist())

            self._target_mask = dilation.Execute(self._target_mask)

            # Update target voxels
            self._target_voxels = np2sitk.sitk_mask_to_linear_indices(
                self._target_mask, order="numpy"
            )

        if not self._target_voxels.any():
            raise GeometryError("No target found in CST")

        # pre-process CT
        self._cube_hu = self._ct.cube_hu

        # Now obtain the target voxel coordinates
        # TODO: 4D issues? And can this be done more

        self._target_voxel_coordinates = np2sitk.linear_indices_to_grid_coordinates(
            self._target_voxels, self._ct.grid, index_type="numpy"
        ).T

        # Multiply with target mask to ignore densities outside of target
        apply_mask = sitk.MaskImageFilter()
        self._cube_hu = apply_mask.Execute(self._cube_hu, self._patient_mask)

    @abstractmethod
    def _generate_source_geometry(self):
        """Generate the source geometry for the STF."""

    def _assign_parameters_from_pln(self, pln):
        """Assign parameters from the plan prop_stf to the generator."""

        pln = validate_pln(pln)

        self._radiation_mode = pln.radiation_mode

        # pln.prop_stf is a dictionary. For every key we try to assign the value
        # If the key is not existing as an attribute, we throw a warning
        for key in pln.prop_stf:
            try:
                setattr(self, key, pln.prop_stf[key])
            except AttributeError:
                warnings.warn(f"Attribute {key} not existing in {self.name} generator")

        self.machine = pln.machine

        if "mult_scen" in pln:
            self.mult_scen = (
                pln.mult_scen
            )  # Validation will be performed as soon as we have the ct
        elif self.mult_scen is None:
            self.mult_scen = "nomScen"
        else:
            pass

        # TODO: self.bio_param = pln.bio_param



"""Module containing treatment planning problem definitions."""

from ._optiprob import NonLinearPlanningProblem, PlanningProblem
from ._nonlin_fluence import NonLinearFluencePlanningProblem
from ._factory import get_available_problems, get_problem, get_problem_from_pln, register_problem

register_problem(NonLinearFluencePlanningProblem)

__all__ = [
    "NonLinearFluencePlanningProblem",
    "NonLinearPlanningProblem",
    "PlanningProblem",
    "get_available_problems",
    "get_problem",
    "get_problem_from_pln",
]



import pytest
import numpy as np
import SimpleITK as sitk

from pyRadPlan.raytracer import RayTracerBase, RayTracerSiddon


@pytest.fixture
def sample_cube():
    cubes = np.random.rand(*(3, 5, 7))  # 3x3x3 cube of ones
    image = sitk.GetImageFromArray(cubes)
    image.SetSpacing([1, 1, 1])
    return image


def test_raytracer_init_siddon(sample_cube):
    raytracer = RayTracerSiddon(sample_cube)
    assert isinstance(raytracer, RayTracerBase)
    assert isinstance(raytracer, RayTracerSiddon)
    assert isinstance(raytracer.lateral_cut_off, float)
    assert len(raytracer.cubes) == 1

    cube_dim = sample_cube.GetSize()

    assert len(raytracer._x_planes) == cube_dim[0] + 1
    assert len(raytracer._y_planes) == cube_dim[1] + 1
    assert len(raytracer._z_planes) == cube_dim[2] + 1


def test_raytracer_trace_single_ray(sample_cube):
    raytracer = RayTracerSiddon(sample_cube)

    isocenter = sample_cube.TransformIndexToPhysicalPoint([3, 2, 1])

    source_points = np.array([0, -5, 0]).astype(float)
    target_points = np.array([0, 5, 0]).astype(float)

    alpha, l, rho, d12, ix = raytracer.trace_ray(isocenter, source_points, target_points)

    assert len(rho) == len(raytracer.cubes)

    # The ray should go through the middle of the cube in y
    cube_np = sitk.GetArrayViewFromImage(sample_cube)
    rho_expected = cube_np.ravel(order="F")[ix]
    assert np.allclose(rho[0], rho_expected)
    assert np.isclose(d12, np.sqrt(np.sum((target_points - source_points) ** 2)))
    assert np.allclose(l, sample_cube.GetSpacing()[1] * np.ones_like(l))  # Spacing is one


def test_raytracer_trace_multiple_rays(sample_cube):
    raytracer = RayTracerSiddon(sample_cube)

    isocenter = sample_cube.TransformIndexToPhysicalPoint([3, 2, 1])

    source_points = np.array([[0, -5, 0], [0, -5, 0]]).astype(float)
    target_points = np.array([[0, 5, 0], [2, 5, 0]]).astype(float)

    alpha, l, rho, d12, ix = raytracer.trace_rays(isocenter, source_points, target_points)

    assert len(rho) == len(raytracer.cubes)

    cube_np = sitk.GetArrayViewFromImage(sample_cube)
    rho_expected = cube_np.ravel(order="F")[ix].reshape(rho[0].shape)
    rho_expected[ix < 0] = 0.0
    rho[0][np.isnan(rho[0])] = 0.0
    assert np.allclose(rho[0], rho_expected)


def test_raytracer_trace_multiple_cubes(sample_cube):
    raytracer = RayTracerSiddon([sample_cube, sample_cube])

    isocenter = sample_cube.TransformIndexToPhysicalPoint([3, 2, 1])

    source_points = np.array([[0, -5, 0], [0, -5, 0]]).astype(float)
    target_points = np.array([[0, 5, 0], [2, 5, 0]]).astype(float)

    alpha, l, rho, d12, ix = raytracer.trace_rays(isocenter, source_points, target_points)

    assert len(rho) == len(raytracer.cubes)

    cube_np = sitk.GetArrayViewFromImage(sample_cube)
    rho_expected = cube_np.ravel(order="F")[ix].reshape(rho[0].shape)
    rho_expected[ix < 0] = 0.0
    rho[0][np.isnan(rho[0])] = 0.0
    rho[1][np.isnan(rho[1])] = 0.0
    assert np.allclose(rho[0], rho_expected)
    assert np.allclose(rho[1], rho_expected)


def test_raytracer_ray_does_not_hit(sample_cube):
    raytracer = RayTracerSiddon(sample_cube)

    isocenter = sample_cube.TransformIndexToPhysicalPoint([3, 2, 1])

    source_point = np.array([100, -5, 100]).astype(float)
    target_point = np.array([100, 5, 100]).astype(float)

    alpha, l, rho, d12, ix = raytracer.trace_ray(isocenter, source_point, target_point)

    assert alpha.size == 0
    assert l.size == 0
    assert rho[0].size == 0
    assert np.isclose(d12, np.sqrt(np.sum((target_point - source_point) ** 2)))
    assert ix.size == 0



"""Dose uniformity."""

from math import sqrt
from numba import njit
from numpy.typing import NDArray

from ._objective import Objective


class DoseUniformity(Objective):
    """Uniformity (minimize standard deviation) objective."""

    name = "Dose Uniformity"

    def compute_objective(self, values):
        return _compute_objective(values)

    def compute_gradient(self, values):
        return _compute_gradient(values)


@njit
def _compute_objective(dose: NDArray):
    return sqrt(len(dose) / (len(dose) - 1)) * dose.std()


@njit
def _compute_gradient(dose: NDArray):
    grad = dose - dose.mean()
    std_val = dose.std()
    if std_val > 0.0:
        grad /= sqrt((len(dose) - 1) * len(dose)) * std_val
    else:
        grad.fill(0.0)

    return grad



"""Beamlet datamodels for particle and photon beamlets."""

from typing import Any
from pydantic import Field, field_serializer, SerializerFunctionWrapHandler, FieldSerializationInfo
import numpy as np
from pyRadPlan.stf._rangeshifter import RangeShifter
from pyRadPlan.core import PyRadPlanBaseModel

# TODO: We need to figure out how we can do nested validation in pydantic and then differentiate
# between photon and particle beamlets. For now, we will use the Beamlet class for both.


class Beamlet(PyRadPlanBaseModel):
    """
    A class representing a single beamlet.

    This class extends PyRadPlanBaseModel (pydantic) and provides functionality to
    handle the bemamlet information, including properties like
    energy & monitor units.

    Attributes
    ----------
    energy : float
        The energy value for the beamlet
    num_particles_per_mu : float
        The number of particles per monitor unit
    relative_fluence : float
        The fluence of this beamlet relative to the central primary fluence.
        For example, due to a non-uniform primary fluence
    weight : float
        The applied fluence weight of the beamlet
    min_mu : float
        The minimum monitor unit
    max_mu : float
        The maximum monitor unit
    range_shifter : RangeShifter
        The range shifter applied for the beamlet.
    focus_ix : int
        The focus index identifying the focus setting for the beamlet.
    """

    energy: float
    num_particles_per_mu: float = Field(alias="numParticlesPerMU", default=1.0e6)
    min_mu: float = Field(alias="minMU", default=0.0)
    max_mu: float = Field(alias="maxMU", default=float("inf"))
    relative_fluence: float = Field(default=1.0)
    weight: float = Field(default=1.0)
    range_shifter: RangeShifter = Field(default_factory=RangeShifter)
    focus_ix: int = Field(default=0)

    @field_serializer(
        "energy",
        "num_particles_per_mu",
        "min_mu",
        "max_mu",
        "relative_fluence",
        "weight",
        "focus_ix",
        mode="wrap",
    )
    def field_typing(
        self, v: Any, handler: SerializerFunctionWrapHandler, info: FieldSerializationInfo
    ) -> Any:
        """Ensure correct serialization in various contexts."""

        context = info.context
        if context and context.get("matRad") == "mat-file":
            if info.field_name == "focus_ix":
                return np.float64(v + 1)  # Increment focus_ix by 1 for MATLAB/matRad
            return np.float64(v)  # Ensure double for MATLAB/matRad

        return handler(v, info)


# Do not use yet


class IonSpot(Beamlet):
    """
    A class representing a single beamlet.

    This class extends PyRadPlanBaseModel (pydantic) and provides functionality to
    handle the beamlet information specific to particles, including properties like
    range shifter and focus index.

    Attributes
    ----------
    range_shifter : RangeShifter
        The range shifter applied for the beamlet.
    focus_ix : int
        The focus index identifying the focus setting for the beamlet.
    """


class PhotonBixel(Beamlet):
    """
    A class representing a single photon beamlet.

    This class extends PyRadPlanBaseModel (pydantic) and provides functionality to
    handle the beamlet information for photons. Mainly the relative fluence of the beamlet /
    bixel due to its lateral position is stored in here.

    Attributes
    ----------
    relative_fluence : float
        The fluence of this beamlet relative to the central primary fluence.
    """


# Example
if __name__ == "__main__":
    testbeamlet = Beamlet(energy=100)



"""Contains the dij class as a (collection of) influence matrices."""

from typing import Any, Union, Annotated, Optional, cast
from pydantic import (
    Field,
    field_validator,
    ValidationInfo,
    computed_field,
    field_serializer,
    SerializationInfo,
    SerializerFunctionWrapHandler,
)

import numpy as np
from numpydantic import NDArray
import SimpleITK as sitk
import scipy.sparse as sp

from pyRadPlan.core import Grid
from pyRadPlan.core import PyRadPlanBaseModel
from pyRadPlan.util import swap_orientation_sparse_matrix


class Dij(PyRadPlanBaseModel):
    """
    Collection of Dose (or other quantity) Influence Matrices.

    Attributes
    ----------
    resolution : dict[str, Any]
        Voxel resolution in each dimension ('x', 'y', 'z').
    physical_dose : scipy.sparse.sparray
        Physical dose matrix.
    total_num_of_bixels : int
        Total number of bixels in the matrix.
    num_of_voxels : int
        Total number of voxels in the matrix.
    """

    dose_grid: Annotated[Grid, Field(default=None)]
    ct_grid: Annotated[Grid, Field(default=None)]

    physical_dose: Annotated[NDArray, Field(default=None)]
    let_dose: Annotated[NDArray, Field(default=None, alias="mLETDose")]
    alpha_dose: Annotated[NDArray, Field(default=None)]
    sqrt_beta_dose: Annotated[NDArray, Field(default=None)]

    num_of_beams: Annotated[int, Field(default=None)]

    bixel_num: Annotated[NDArray, Field(default=None)]
    ray_num: Annotated[NDArray, Field(default=None)]
    beam_num: Annotated[NDArray, Field(default=None)]

    rad_depth_cubes: Optional[list[NDArray]] = Field(default=None)

    @computed_field
    @property
    def total_num_of_bixels(self) -> int:
        """Number of bixels / beamlets in the dose influence matrix."""
        return int(self.bixel_num.size)

    @computed_field
    @property
    def num_of_voxels(self) -> int:
        """Number of voxels in the dose influence matrix."""
        return self.physical_dose.flat[0].shape[0]

    @computed_field
    @property
    def quantities(self) -> list[str]:
        """Name of available uantities matrices."""
        potential_quantities = ["physical_dose", "let_dose", "alpha_dose", "sqrt_beta_dose"]
        return [q for q in potential_quantities if getattr(self, q) is not None]

    @field_validator("physical_dose", "let_dose", "alpha_dose", "sqrt_beta_dose", mode="before")
    @classmethod
    def validate_matrices(cls, v: Any, info: ValidationInfo) -> np.ndarray:
        """
        Validate the physical dose matrix.

        Raises
        ------
            ValueError: if physical dose is not a 2D numpy array.
        """

        if v is None:
            return v

        if isinstance(v, (sp.spmatrix, sp.sparray)) or (
            isinstance(v, np.ndarray) and v.dtype != np.dtype(object)
        ):
            tmp = np.empty((1,), dtype=object)
            tmp[0] = v
            v = tmp

        if isinstance(v, list):
            v = np.asarray(v, dtype=object)

        for i in range(v.size):
            if v.flat[i] is not None:
                mat = v.flat[i]
                if not isinstance(mat, (sp.spmatrix, sp.sparray, np.ndarray)) or not np.issubdtype(
                    mat.dtype, np.number
                ):
                    raise ValueError(f"{info.field_name} must be a numeric array.")
                if not mat.ndim == 2:
                    raise ValueError(f"{info.field_name} must be a 2D array.")
                if not mat.shape == mat.shape:
                    raise ValueError(f"{info.field_name} must have consistent number of voxels.")

                if mat.shape[0] != info.data["dose_grid"].num_voxels:
                    raise ValueError(f"{info.field_name} shape inconsistent with ct grid")

        if info.context and "from_matRad" in info.context and info.context["from_matRad"]:
            if v is not None:
                for i in range(v.size):
                    shape = (
                        int(info.data["dose_grid"].dimensions[2]),
                        int(info.data["dose_grid"].dimensions[0]),
                        int(info.data["dose_grid"].dimensions[1]),
                    )
                    v.flat[i] = swap_orientation_sparse_matrix(
                        v.flat[i],
                        shape,
                        (1, 2),  # (65, 100, 100) example
                    )
                    if v.flat[i] is not None and not isinstance(v.flat[i], sp.csc_matrix):
                        v.flat[i] = sp.csc_matrix(v.flat[i])
            else:
                v = np.array([0])
            return [[v]]  # TODO

        return v

    @field_validator("dose_grid", "ct_grid", mode="before")
    @classmethod
    def validate_grid(cls, grid: Union[Grid, dict], info: ValidationInfo) -> Union[Grid, dict]:
        """
        Validate grid dictionaries.

        Raises
        ------
            ValueError:
        """
        # Check if it is a dictionary and then try to create a Grid object
        if isinstance(grid, dict):
            if info.context and "from_matRad" in info.context and info.context["from_matRad"]:
                grid["dimensions"] = np.array(
                    [grid["dimensions"][1], grid["dimensions"][0], grid["dimensions"][2]]
                )
                # TODO: might swap offset and resolution
                grid = Grid.model_validate(grid)
            else:
                grid = Grid.model_validate(grid)
        return grid

    @field_validator("beam_num", mode="before")
    @classmethod
    def validate_unique_indices_in_beam_num(
        cls, v: np.ndarray, info: ValidationInfo
    ) -> np.ndarray:
        """
        Validate the number of unique indices in beam_num.

        Raises
        ------
            ValueError: Number of unique indices does not match number of beams.
        """
        num_of_beams = info.data["num_of_beams"]
        if len(np.unique(v)) != num_of_beams:
            raise ValueError(
                "Number of unique indices in beam_num does not match number of beams."
            )
        return v

    @field_validator("beam_num", "ray_num", "bixel_num", mode="after")
    @classmethod
    def validate_numbering_arrays(cls, v: np.ndarray, info: ValidationInfo) -> np.ndarray:
        """
        Validate the numbering arrays.

        Raises
        ------
            ValueError: inconsistent numbering arrays.
        """
        # Check if the numbering arrays have the correct shape
        if info.data.get("physical_dose") is not None:
            dij_matrices = cast(np.ndarray, info.data["physical_dose"])
            for i in range(dij_matrices.size):
                if dij_matrices.flat[i] is not None:
                    mat = cast(Union[sp.spmatrix, sp.sparray, np.ndarray], dij_matrices.flat[i])

                    bix_num = mat.shape[1]

                    if v.ndim != 1:
                        raise ValueError("Numbering arrays must be 1-dimensional")

                    if v.size != bix_num:
                        raise ValueError(
                            "Numbering arrays shape inconsistent with number of bixels"
                        )

        if info.context and "from_matRad" in info.context and info.context["from_matRad"]:
            v -= 1
        return v

    # Serialization
    @field_serializer("dose_grid", "ct_grid", mode="wrap")
    def grid_serializer(
        self, value: Grid, handler: SerializerFunctionWrapHandler, info: SerializationInfo
    ) -> dict:
        context = info.context
        if context and context.get("matRad") == "mat-file":
            return value.to_matrad(context=context["matRad"])
        return handler(value, info)

    @field_serializer("physical_dose", "let_dose", "alpha_dose", "sqrt_beta_dose")
    def physical_dose_serializer(self, value: np.ndarray, info: SerializationInfo) -> np.ndarray:
        context = info.context
        if context and context.get("matRad") == "mat-file" and value is not None:
            for i in range(value.size):
                shape = (
                    int(self.dose_grid.dimensions[2]),
                    int(self.dose_grid.dimensions[0]),
                    int(self.dose_grid.dimensions[1]),
                )
                value.flat[i] = swap_orientation_sparse_matrix(
                    value.flat[i],
                    shape,
                    (1, 2),  # (65, 100, 100) example
                )
                if value.flat[i] is not None and not isinstance(value.flat[i], sp.csc_matrix):
                    value.flat[i] = sp.csc_matrix(value.flat[i])
        # return 0 if value is None. savemat() cant handle 'None'
        elif context and context.get("matRad") == "mat-file" and value is None:
            value = np.array([0])
        return value

    @field_serializer("rad_depth_cubes")
    def rad_depth_cubes_serializer(self, value: np.ndarray, info: SerializationInfo) -> np.ndarray:
        context = info.context
        if context and context.get("matRad") == "mat-file" and value is not None:
            # TODO: it might be necessary to rotate the cube!
            return value
        # return 0 if value is None. savemat() cant handle 'None'
        elif context and context.get("matRad") == "mat-file" and value is None:
            value = np.array([0])
        return value

    @field_serializer("bixel_num", "ray_num", "beam_num")
    def numbering_arrays_serializer(
        self, value: np.ndarray, info: SerializationInfo
    ) -> np.ndarray:
        context = info.context
        if context and context.get("matRad") == "mat-file":
            return value.reshape(-1, 1)
        return value

    def to_matrad(self, context: str = "mat-file") -> Any:
        """Convert the Dij to matRad-compatible dictionary."""

        dij_dict = super().to_matrad(context=context)

        return dij_dict

    def get_result_arrays_from_intensity(
        self, intensity: np.ndarray, scenario_index: int = 0
    ) -> dict[str, np.ndarray]:
        """
        Compute result arrays from an intensity vector.

        Parameters
        ----------
        intensity : np.ndarray
            The intensity to apply to the dose influence matrix.
        scenario_index : int
            The scenario index to apply the intensity to.

        Returns
        -------
        dict[str,sitk.Image]
            A dictionary containing the quantity images for each scenario.
        """

        out = {}

        # TODO: implement quantity system to select the corresponding quantities automatically
        if self.physical_dose is not None:
            out["physical_dose"] = self.physical_dose.flat[scenario_index] @ intensity

        if self.let_dose is not None:
            if self.physical_dose is None:
                raise ValueError("Physical dose must be calculated for dose-weighted let")

            indices = out["physical_dose"] > 0.05 * np.max(out["physical_dose"])

            let_dose = self.let_dose.flat[scenario_index] @ intensity
            out["let"] = np.zeros_like(let_dose)
            out["let"][indices] = let_dose[indices] / out["physical_dose"][indices]

        if self.alpha_dose is not None and self.sqrt_beta_dose is not None:
            out["effect"] = (
                self.alpha_dose.flat[scenario_index] @ intensity
                + (self.sqrt_beta_dose.flat[scenario_index] @ intensity) ** 2
            )

        return out

    def compute_result_dose_grid(
        self, intensities: np.ndarray, scenario_index: int = 0
    ) -> dict[str, sitk.Image]:
        """
        Compute results on the dose grid from intensity vector.

        Parameters
        ----------
        intensity : np.ndarray
            The intensity to apply to the dose influence matrix.
        scenario_index : int
            The scenario index to apply the intensity to.

        Returns
        -------
        dict[str,sitk.Image]
            A dictionary containing the quantity images for each scenario.
        """

        out = self.get_result_arrays_from_intensity(intensities, scenario_index=scenario_index)
        # Create a sitk image for each scenario

        for key, value in out.items():
            # Create a sitk image for each scenario
            out[key] = sitk.GetImageFromArray(value.reshape(self.dose_grid.dimensions[::-1]))
            out[key].SetOrigin(self.dose_grid.origin)
            out[key].SetSpacing(self.dose_grid.resolution_vector)
            out[key].SetDirection(self.dose_grid.direction.ravel())

        return out

    def compute_result_ct_grid(
        self, intensities: np.ndarray, scenario_index: int = 0
    ) -> dict[str, sitk.Image]:
        """
        Compute results on the CT grid from intensity vector.

        Parameters
        ----------
        intensity : np.ndarray
            The intensity to apply to the dose influence matrix.
        scenario_index : int
            The scenario index to apply the intensity to.

        Returns
        -------
        dict[str,sitk.Image]
            A dictionary containing the quantity images for each scenario.
        """

        out = self.compute_result_dose_grid(intensities, scenario_index=scenario_index)
        # Create a sitk image for each scenario

        for key, value in out.items():
            # Create a sitk image for each scenario
            resampler = sitk.ResampleImageFilter()
            resampler.SetInterpolator(sitk.sitkLinear)
            resampler.SetOutputDirection(self.ct_grid.direction.ravel())
            resampler.SetOutputOrigin(self.ct_grid.origin)
            resampler.SetOutputSpacing(self.ct_grid.resolution_vector)
            resampler.SetSize(self.ct_grid.dimensions)
            out[key] = resampler.Execute(value)

        return out


def create_dij(data: Union[dict[str, Any], Dij, None] = None, **kwargs) -> Dij:
    """
    Create a Dij object from raw data or keyword arguments.

    Parameters
    ----------
    data : Union[dict[str, Any], Dij, None]
        Dictionary containing the data to create the Dij object.
    **kwargs
        Arbitrary keyword arguments.

    Returns
    -------
    Dij
        A Dij object.
    """

    if data:
        # If data is already a Dij object, return it directly
        if isinstance(data, Dij):
            return data

        if "beamNum" in data and np.min(data["beamNum"]) != 0:
            # add context when from matRad
            context = {"from_matRad": True}
        else:
            context = {"from_matRad": False}
        return Dij.model_validate(data, context=context)

    return Dij(**kwargs)


def validate_dij(dij: Union[dict[str, Any], Dij, None] = None, **kwargs) -> Dij:
    """
    Validate and creates a Dij object.

    Synonym to create_dij but should be used in validation context.

    Parameters
    ----------
    dij : Union[dict[str, Any], Dij, None], optional
        Dictionary containing the data to create the Dij object, by default None.
    **kwargs
        Arbitrary keyword arguments.

    Returns
    -------
    Dij
        A validated Dij object.

    """
    return create_dij(dij, **kwargs)



from numba import njit, prange
import numpy as np


@njit(parallel=True, nogil=True, cache=True)
def fast_spatial_circle_lookup(
    mesh_x: np.ndarray, mesh_z: np.ndarray, lookup_pos: np.ndarray, radius: float
) -> np.ndarray:
    """Lookup all points within a circle in a meshgrid.

    Parameters
    ----------
    mesh_x : np.ndarray
        The x-coordinates of the meshgrid (NxN)
    mesh_z : np.ndarray
        The y-coordinates of the meshgrid (NxN)
    lookup_pos : np.ndarray
        the reference positions to lookup (Mx3)
    radius : float
        The radius of the circle.

    Returns
    -------
    np.ndarray
        The indices of the points within the circle.
    """
    candidate_ray_mx = np.full(mesh_x.shape, False, dtype=np.bool_)
    for i in prange(lookup_pos.shape[0]):
        ix = (mesh_x.ravel() - lookup_pos[i, 0]) ** 2 + (
            mesh_z.ravel() - lookup_pos[i, 2]
        ) ** 2 <= radius**2
        candidate_ray_mx.ravel()[ix] = 1
    return candidate_ray_mx


@njit
def _fast_compute_plane_alphas(
    dim_min: np.ndarray,
    dim_max: np.ndarray,
    planes: np.ndarray,
    source: np.ndarray,
    ray: np.ndarray,
) -> np.ndarray:
    num_rays = source.shape[0]
    dim_min = np.atleast_2d(dim_min).T
    dim_max = np.atleast_2d(dim_max).T

    # plane_grid = np.tile(planes, (num_rays, 1)) # not numba compatible
    plane_grid = np.repeat(planes, num_rays).reshape((-1, num_rays)).T
    plane_ix = np.atleast_2d(np.arange(0, len(planes)))

    # set invalid indices to NaN
    ix = (
        (plane_ix < dim_min)
        | (plane_ix > dim_max)
        | ((plane_ix == dim_min) & (plane_ix == dim_max))
        | np.isnan(dim_min)
        | np.isnan(dim_max)
    )
    # np.isnan()

    plane_grid.ravel()[ix.ravel()] = np.nan

    # repeat.reshape.T replaces np.tile for numba compatibility
    source_vec_t = np.repeat(source, plane_grid.shape[1]).reshape((-1, plane_grid.shape[1]))
    ray_vec_t = np.repeat(ray, plane_grid.shape[1]).reshape((-1, plane_grid.shape[1]))

    alphas = (plane_grid - source_vec_t) / ray_vec_t

    return alphas


@njit
def _fast_compute_all_alphas(
    source_points: np.ndarray,
    ray_vecs: np.ndarray,
    x_planes: np.ndarray,
    y_planes: np.ndarray,
    z_planes: np.ndarray,
    resolution: np.ndarray,
    precision: np.dtype,
) -> np.ndarray:
    num_rays = ray_vecs.shape[0]

    a_x_min = np.zeros(num_rays, dtype=precision)
    a_x_max = np.ones(num_rays, dtype=precision)
    a_x_ix = np.nonzero(ray_vecs[:, 0])
    if len(a_x_ix[0]) > 0:
        tmp = (
            np.column_stack(
                (
                    x_planes[0] - source_points[a_x_ix[0], 0],
                    x_planes[-1] - source_points[a_x_ix[0], 0],
                )
            )
            / ray_vecs[a_x_ix[0], 0, None]
        )
        for j in range(tmp.shape[0]):
            a_x_min[a_x_ix[0][j]] = np.nanmin(tmp[j])
            a_x_max[a_x_ix[0][j]] = np.nanmax(tmp[j])

    a_y_min = np.zeros(num_rays, dtype=precision)
    a_y_max = np.ones(num_rays, dtype=precision)
    a_y_ix = np.nonzero(ray_vecs[:, 1])
    if len(a_y_ix[0]) > 0:
        tmp = (
            np.column_stack(
                (
                    y_planes[0] - source_points[a_y_ix[0], 1],
                    y_planes[-1] - source_points[a_y_ix[0], 1],
                )
            )
            / ray_vecs[a_y_ix[0], 1, None]
        )
        for j in range(tmp.shape[0]):
            a_y_min[a_y_ix[0][j]] = np.nanmin(tmp[j])
            a_y_max[a_y_ix[0][j]] = np.nanmax(tmp[j])

    a_z_min = np.zeros(num_rays, dtype=precision)
    a_z_max = np.ones(num_rays, dtype=precision)
    a_z_ix = np.nonzero(ray_vecs[:, 2])
    if len(a_z_ix[0]) > 0:
        tmp = (
            np.column_stack(
                (
                    z_planes[0] - source_points[a_z_ix[0], 2],
                    z_planes[-1] - source_points[a_z_ix[0], 2],
                )
            )
            / ray_vecs[a_z_ix[0], 2, None]
        )
        for j in range(tmp.shape[0]):
            a_z_min[a_z_ix[0][j]] = np.nanmin(tmp[j])
            a_z_max[a_z_ix[0][j]] = np.nanmax(tmp[j])

    # Compute alpha_limits
    alpha_min_values = np.full(num_rays, np.nan, dtype=precision)
    alpha_max_values = np.full(num_rays, np.nan, dtype=precision)

    for i in range(num_rays):
        alpha_min_values[i] = np.nanmax(
            np.array([0, a_x_min[i], a_y_min[i], a_z_min[i]], dtype=precision)
        )
        alpha_max_values[i] = np.nanmin(
            np.array([1, a_x_max[i], a_y_max[i], a_z_max[i]], dtype=precision)
        )

    alpha_limits = np.column_stack((alpha_min_values, alpha_max_values))

    # Determine the direction of ray movement
    # ix_tp_bigger_sp = self._target_points > source_points
    # ix_tp_smaller_sp = self._target_points < source_points
    ix_tp_bigger_sp = ray_vecs > 0.0
    ix_tp_smaller_sp = ray_vecs < 0.0

    for d in range(3):  # Loop through x, y, z dimensions
        alpha_tmp = np.full((num_rays, 2), np.nan, dtype=precision)
        alpha_tmp[ix_tp_bigger_sp[:, d], 0] = alpha_limits[ix_tp_bigger_sp[:, d], 0]
        alpha_tmp[ix_tp_smaller_sp[:, d], 0] = alpha_limits[ix_tp_smaller_sp[:, d], 1]
        alpha_tmp[ix_tp_bigger_sp[:, d], 1] = alpha_limits[ix_tp_bigger_sp[:, d], 1]
        alpha_tmp[ix_tp_smaller_sp[:, d], 1] = alpha_limits[ix_tp_smaller_sp[:, d], 0]

        # Compute min and max indices (i_min, i_max, j_min, j_max, k_min, k_max)
        if d == 0:  # x dimension
            planes = x_planes
            res_d = resolution[0]
        elif d == 1:  # y dimension
            planes = y_planes
            res_d = resolution[1]
        else:  # z dimension
            planes = z_planes
            res_d = resolution[2]

        num_planes = planes.size

        dim_min = (
            num_planes
            - (planes[-1] - alpha_tmp[:, 0] * ray_vecs[:, d] - source_points[:, d]) / res_d
            - 1
        )
        dim_max = (source_points[:, d] + alpha_tmp[:, 1] * ray_vecs[:, d] - planes[0]) / res_d
        # Rounding
        dim_min = np.ceil(1 / 1000 * (np.round(1000 * dim_min)))
        dim_max = np.floor(1 / 1000 * (np.round(1000 * dim_max)))

        if d == 0:  # Assigning computed min and max indices to i_min, i_max
            i_min, i_max = dim_min, dim_max
        elif d == 1:  # Assigning computed min and max indices to j_min, j_max
            j_min, j_max = dim_min, dim_max
        else:  # Assigning computed min and max indices to k_min, k_max
            k_min, k_max = dim_min, dim_max

    # Compute alphas for each plane and merge parametric sets
    alpha_x = _fast_compute_plane_alphas(
        i_min,
        i_max,
        x_planes,
        source_points[:, 0],
        ray_vecs[:, 0],
    )
    alpha_y = _fast_compute_plane_alphas(
        j_min,
        j_max,
        y_planes,
        source_points[:, 1],
        ray_vecs[:, 1],
    )
    alpha_z = _fast_compute_plane_alphas(
        k_min,
        k_max,
        z_planes,
        source_points[:, 2],
        ray_vecs[:, 2],
    )

    alphas = np.concatenate((alpha_limits, alpha_x, alpha_y, alpha_z), axis=1)

    # Vectorized unique operation across rows
    # Sort alphas row-wise and remove duplicates
    alphas = np.sort(alphas, axis=1)  # Sort each row ascendingly
    mask = np.diff(alphas, axis=1) == 0  # Identify duplicates
    alphas[:, 1:][mask] = np.nan  # Replace duplicates with NaN
    alphas = np.sort(alphas, axis=1)

    # Size Reduction
    max_num_columns = np.max(np.sum(~np.isnan(alphas), axis=1))
    alphas = alphas[:, :max_num_columns]

    return alphas



"""Basic Model for all pyRadPlan Datastructures."""

from typing import Any, Union
import numpy as np
from pydantic import (
    AliasGenerator,
    BaseModel,
    ConfigDict,
)
from pydantic.alias_generators import to_camel
from copy import deepcopy


class PyRadPlanBaseModel(BaseModel):
    """
    Base class for all pyRadPlan data structures.

    Especially useful for structures that should be matRad compatible.
    Extends Pydantic's BaseModel to use pydantic validation and serialization.

    Attributes
    ----------
    model_config : ConfigDict
        Configuration for the model, including alias generation, population by
        name, arbitrary types allowed, assignment validation, and attribute
        creation from dictionary.
    """

    model_config = ConfigDict(
        alias_generator=AliasGenerator(alias=to_camel),
        populate_by_name=True,  # Allows both snake_case and camelCase attributes
        arbitrary_types_allowed=True,  # Allows arbitrary types in the model (will be casted)
        validate_assignment=True,  # Validate assignment of values to fields
        # (not only during construction)
        from_attributes=True,  # Allows to create a model from a dictionary
    )

    def __eq__(self, other: Any) -> bool:
        """
        Specialized __eq__ method to compare two pyRadPlanBaseModel instances.

        It first tries to compare the instances using the super().__eq__ method.
        If this fails, it compares the dictionaries. This is due to some issues
        comparing numpy arrays within the models.
        """
        try:
            return super().__eq__(other)
        except ValueError:
            if self.__dict__.keys() != other.__dict__.keys():
                return False
            stack = [(self.__dict__, other.__dict__)]
            while stack:
                dict_a, dict_b = stack.pop()
                if dict_a.keys() != dict_b.keys():
                    return False
                for key in dict_a:
                    if isinstance(dict_a[key], dict) and isinstance(dict_b[key], dict):
                        stack.append((dict_a[key], dict_b[key]))
                    elif isinstance(dict_a[key], np.ndarray) and isinstance(
                        dict_b[key], np.ndarray
                    ):
                        if not np.array_equal(dict_a[key], dict_b[key]):
                            return False
                    elif dict_a[key] != dict_b[key]:
                        return False
            return True

    def __ne__(self, other: Any) -> bool:
        """
        Specialized __ne__ method to compare two PyRadPlanBaseModel instances.

        This method returns the negation of the __eq__ method.
        """
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        else:
            return True

    def to_matrad(self, context: Union[str, dict] = "mat-file") -> Any:
        """
        Perform matRad compatible serialization.

        Parameters
        ----------
        context : str, optional
            The context in which the datastructure should be serialized,
            by default 'mat-file'.

        Returns
        -------
        Any
            A datastructre compatible with matRad in the given context

        Notes
        -----
        Currently, the only supported context is 'mat-file'. In the future,
        this could be extended to support other contexts, such as direct
        calling via the matlab engine or oct2py.
        """
        self_copy = deepcopy(self)
        if isinstance(context, dict):
            if "matRad" not in context:
                context.update({"matRad": "mat-file"})
        else:
            context = {"matRad": context}

        if context["matRad"] != "mat-file":
            raise ValueError(f"Context {context} not supported")

        # Standard is a model_dump using above alias and context information
        return self_copy.model_dump(by_alias=True, context=context)



from typing import Dict
import pytest

from pyRadPlan.plan import IonPlan, PhotonPlan
from pyRadPlan.stf import SteeringInformation
from pyRadPlan.dij import Dij
from pyRadPlan.ct import CT
from pyRadPlan.cst import StructureSet


from pyRadPlan.plan import validate_pln
from pyRadPlan.stf import validate_stf
from pyRadPlan.dij import validate_dij
from pyRadPlan.ct import validate_ct
from pyRadPlan.cst import validate_cst


@pytest.fixture
def test_data_photons(test_data_photons_raw):
    tmp = test_data_photons_raw

    pln = validate_pln(tmp["pln"])
    pln.prop_dose_calc["dosimetric_lateral_cutoff"] = 0.995
    pln.prop_dose_calc["lateral_model"] = "single"

    ct = validate_ct(tmp["ct"])
    cst = validate_cst(tmp["cst"], ct=ct)
    dij_matRad = validate_dij(tmp["dij"])
    stf = validate_stf(tmp["stf"])

    result = tmp["resultGUI"]

    return pln, ct, cst, stf, dij_matRad, result


@pytest.fixture
def test_data_protons(test_data_protons_raw):
    tmp = test_data_protons_raw

    pln = validate_pln(tmp["pln"])
    pln.prop_dose_calc["dosimetric_lateral_cutoff"] = 0.995
    pln.prop_dose_calc["lateral_model"] = "single"

    ct = validate_ct(tmp["ct"])
    cst = validate_cst(tmp["cst"], ct=ct)
    dij_matRad = validate_dij(tmp["dij"])
    stf = validate_stf(tmp["stf"])

    result = tmp["resultGUI"]

    return pln, ct, cst, stf, dij_matRad, result


@pytest.fixture
def test_data_helium(test_data_helium_raw):
    tmp = test_data_helium_raw

    pln = validate_pln(tmp["pln"])
    pln.prop_dose_calc["dosimetric_lateral_cutoff"] = 0.995
    pln.prop_dose_calc["lateral_model"] = "single"

    ct = validate_ct(tmp["ct"])
    cst = validate_cst(tmp["cst"], ct=ct)
    dij_matRad = validate_dij(tmp["dij"])
    stf = validate_stf(tmp["stf"])

    result = tmp["resultGUI"]

    return pln, ct, cst, stf, dij_matRad, result


@pytest.fixture
def test_data_carbon(test_data_carbon_raw):
    tmp = test_data_carbon_raw

    pln = validate_pln(tmp["pln"])
    pln.prop_dose_calc["dosimetric_lateral_cutoff"] = 0.995
    pln.prop_dose_calc["lateral_model"] = "single"

    ct = validate_ct(tmp["ct"])
    cst = validate_cst(tmp["cst"], ct=ct)
    dij_matRad = validate_dij(tmp["dij"])
    stf = validate_stf(tmp["stf"])

    result = tmp["resultGUI"]

    return pln, ct, cst, stf, dij_matRad, result


def test_validate_protons(test_data_protons):
    pln_p, ct_p, cst_p, stf_p, dij_p, result_p = test_data_protons
    assert isinstance(pln_p, IonPlan)
    assert isinstance(ct_p, CT)
    assert isinstance(cst_p, StructureSet)
    assert isinstance(stf_p, SteeringInformation)
    assert isinstance(dij_p, Dij)
    assert isinstance(result_p, Dict)


def test_validate_helium(test_data_helium):
    pln_h, ct_h, cst_h, stf_h, dij_h, result_h = test_data_helium
    assert isinstance(pln_h, IonPlan)
    assert isinstance(ct_h, CT)
    assert isinstance(cst_h, StructureSet)
    assert isinstance(stf_h, SteeringInformation)
    assert isinstance(dij_h, Dij)
    assert isinstance(result_h, Dict)


def test_validate_carbon(test_data_carbon):
    pln_c, ct_c, cst_c, stf_c, dij_c, result_c = test_data_carbon
    assert isinstance(pln_c, IonPlan)
    assert isinstance(ct_c, CT)
    assert isinstance(cst_c, StructureSet)
    assert isinstance(stf_c, SteeringInformation)
    assert isinstance(dij_c, Dij)
    assert isinstance(result_c, Dict)


def test_validate_photons(test_data_photons):
    pln_p, ct_p, cst_p, stf_p, dij_p, result_p = test_data_photons
    assert isinstance(pln_p, PhotonPlan)
    assert isinstance(ct_p, CT)
    assert isinstance(cst_p, StructureSet)
    assert isinstance(stf_p, SteeringInformation)
    assert isinstance(dij_p, Dij)
    assert isinstance(result_p, Dict)



"""Module for Dij Datamodel and related functions."""

from pyRadPlan.dij._dij import Dij, create_dij, validate_dij
from pyRadPlan.dij._compose_beam_dijs import compose_beam_dijs

__all__ = ["Dij", "create_dij", "validate_dij", "compose_beam_dijs"]



import os
import sys
from typing import Union

if sys.version_info < (3, 10):
    import importlib_resources as resources  # Backport for older versions
else:
    from importlib import resources  # Standard from Python 3.9+

from pymatreader import read_mat
from pyRadPlan.machines import Machine, validate_machine


def load_machine(filename: str) -> Machine:
    if filename.endswith(".mat"):
        machine_dict = load_machine_from_mat(filename)
    else:
        raise ValueError("Unsupported file type for machine loading:", filename)

    return validate_machine(machine_dict)


def load_machine_from_mat(filename: Union[os.PathLike, str]) -> dict:
    try:
        machine = read_mat(filename)
        return machine["machine"]
    except FileNotFoundError as exc:
        raise FileNotFoundError("Could not find the following machine file:", filename) from exc
    except KeyError as exc:
        raise FileNotFoundError(
            "Could not find the machine key in the following file:", filename
        ) from exc


def load_from_name(radiation_mode: str, machine_name: str) -> Machine:
    possible_endings = [".mat"]

    # Search resources
    # TODO: allow customization of search path
    search_paths = [resources.files("pyRadPlan.data.machines")]

    for ending in possible_endings:
        composed_filename = radiation_mode + "_" + machine_name + ending

        # We traverse all search paths
        for current_path in search_paths:
            filename = current_path.joinpath(composed_filename)
            if filename.exists():
                return load_machine(filename.as_posix())

    raise FileNotFoundError("Could not find the following machine file:", composed_filename)


if __name__ == "__main__":
    machine = load_from_name(radiation_mode="photons", machine_name="Generic")
    print(machine)



"""Base Implementation for objective functions."""

from abc import abstractmethod
from typing import ClassVar, Any, Literal, Union, Optional
import logging

from pydantic import computed_field, Field, field_validator, model_validator

from pyRadPlan.core.datamodel import PyRadPlanBaseModel
from pyRadPlan.quantities import get_available_quantities

ParameterType = Union[Literal["reference", "numeric", "relative_volume"], list[str]]

logger = logging.getLogger(__name__)


class ParameterMetadata:
    """
    Parameter Metadata to attach to objective function parameters to
    designate configurability and
    type.

    Parameters
    ----------
    configurable : bool, optional, default=True
        Whether the parameter is intended to be configurable (e.g. a reference dose value for
        optimization)
    kind : str, optional, default='numeric'
        Type/Meaning of the parameter. Options are 'reference' (relates to the input vector, e.g.
        the dose when you optimize on dose. Used for normalization), 'numeric', 'relative_volume'.
        The kind is used in case the parameter needs to pre transformed in context.

    Attributes
    ----------
    configurable : bool, optional, default=True
        Whether the parameter is intended to be configurable (e.g. a reference dose value for
        optimization)
    kind : str, optional, default='numeric'
        Type/Meaning of the parameter. Options are 'reference' (relates to the input vector, e.g.
        the dose when you optimize on dose. Used for normalization), 'numeric', 'relative_volume'.
        The kind is used in case the parameter needs to pre transformed in context.

    Note
    ----
    Intended for use with 'Annotated' when defining an objective attribute.
    """

    configurable: bool
    kind: Optional[ParameterType]

    """Configurable Parameter."""

    def __init__(self, configurable: bool = True, kind: Optional[ParameterType] = "numeric"):
        self.configurable = configurable
        self.kind = kind

    def __repr__(self):
        return f"{self.__class__}({self.__dict__})"


class Objective(PyRadPlanBaseModel):
    """
    Base class for objective functions in the optimization problem.

    Attributes
    ----------
    name : ClassVar[str]
        Name of the objective function.
    has_hessian : ClassVar[bool]
        Whether the objective function has a Hessian implementation.
    priority : float
        Weight/Priority assigned to the objective function.
    quantity : str
        The quantity this objective is connected to (e.g. 'physical_dose', 'RBExDose').
    """

    name: ClassVar[str]
    has_hessian: ClassVar[bool] = False
    priority: float = Field(default=1.0, ge=0.0, alias="penalty")
    quantity: str = Field(default="physical_dose")

    @abstractmethod
    def compute_objective(self, values):
        """Computes the objective function."""

    @abstractmethod
    def compute_gradient(self, values):
        """Computes the objective gradient."""

    def compute_hessian(self, values):
        """Computes the objective Hessian."""
        return

    @computed_field
    @property
    def parameter_names(self) -> list[str]:
        """List[str]: Parameter names."""
        return self._parameter_names()

    @classmethod
    def _parameter_names(cls) -> list[str]:
        """List[str]: Parameter names as classmethod."""
        return [
            name
            for name, info in cls.model_fields.items()
            if any(isinstance(meta, ParameterMetadata) for meta in info.metadata)
        ]

    @computed_field
    @property
    def parameter_types(self) -> list[ParameterType]:
        """List[str]: Parameter types."""
        return self._parameter_types()

    @classmethod
    def _parameter_types(cls) -> list[ParameterType]:
        """List[str]: Parameter types as classmethod."""
        return [
            meta.kind
            for name in cls._parameter_names()
            for meta in cls.model_fields[name].metadata
            if isinstance(meta, ParameterMetadata)
        ]

    @computed_field
    @property
    def parameters(self) -> list[Any]:
        """List[str]: Parameter values."""
        return [getattr(self, name) for name in self.parameter_names]

    @field_validator("quantity")
    @classmethod
    def _validate_quantity(cls, v):
        """Validates the quantity attribute."""
        if v not in get_available_quantities():
            raise ValueError(
                f"Quantity {v} not available. Choose from {get_available_quantities()}"
            )
        return v

    @model_validator(mode="before")
    @classmethod
    def _validate_model(cls, data: Any) -> Any:
        """Pre-validate the input and perform conversions if necessary."""

        # Check if this is a matRad-like objective
        if isinstance(data, dict) and "className" in data:
            data = data.copy()

            # Should we confirm once more we have the correct objective?
            data.pop("className")

            params = data.get("parameters", [])

            # If there are not more than one parameter,
            # it will usually not be in a list so we put it into one
            if not isinstance(params, list):
                params = [params]

            # obtain the parameter names
            param_names = cls._parameter_names()

            if len(params) != len(param_names):
                logger.warning(
                    "Objective '%s' expects %d parameters, but %d were provided.",
                    cls.name,
                    len(param_names),
                    len(params),
                )

            for param in param_names:
                data[param] = params.pop(0)

            data.pop("parameters", None)

        return data



import pytest
import pymatreader
import sys

if sys.version_info < (3, 10):
    import importlib_resources as resources  # Backport for older versions
else:
    from importlib import resources  # Standard from Python 3.9+


@pytest.fixture
def tg119_raw():
    phantom_data_str = resources.files("pyRadPlan.data.phantoms").joinpath("TG119.mat")
    phantom_data = pymatreader.read_mat(phantom_data_str)
    return phantom_data["ct"], phantom_data["cst"]


@pytest.fixture
def test_data_photons_raw():
    path = "test/data/photons_testData.mat"
    tmp = pymatreader.read_mat(path)
    return tmp


@pytest.fixture
def test_data_protons_raw():
    path = "test/data/protons_testData.mat"
    tmp = pymatreader.read_mat(path)
    return tmp


@pytest.fixture
def test_data_helium_raw():
    path = "test/data/helium_testData.mat"
    tmp = pymatreader.read_mat(path)
    return tmp


@pytest.fixture
def test_data_carbon_raw():
    path = "test/data/carbon_testData.mat"
    tmp = pymatreader.read_mat(path)
    return tmp



from ._matlab_file_handler import MatlabFileHandler
from ._patient_loader import load_patient, load_tg119, validate_matrad_patient

__all__ = ["MatlabFileHandler", "load_patient", "load_tg119", "validate_matrad_patient"]



import pytest

import numpy as np
import SimpleITK as sitk
import matplotlib

from pyRadPlan import CT, StructureSet
from pyRadPlan import plot_slice

# Use Agg backend for matplotlib, so plots are not displayed during testing. (even if plt_show=False)
# When running tests locally, you can find the files in
# %appdata%\local\Temp\pytest-of-<username>\pytest-<number>.
matplotlib.use("Agg")


@pytest.fixture
def sample_ct():
    cube_hu = sitk.Image(100, 100, 10, sitk.sitkFloat32)
    return CT(cube_hu=cube_hu)


@pytest.fixture
def sample_cst(sample_ct):
    mask_np = np.zeros((10, 100, 100), dtype=np.uint8)
    mask_np[1:8, 20:80, 20:80] = 1
    mask = sitk.GetImageFromArray(mask_np)
    return StructureSet(
        vois=[{"mask": mask, "voi_type": "TARGET", "name": "test", "ct_image": sample_ct}],
        ct_image=sample_ct,
    )


@pytest.fixture
def sample_overlay():
    rand_image = np.random.rand(10, 100, 100).astype(np.float32)
    overlay = sitk.GetImageFromArray(rand_image)
    return overlay


def test_plot_slice_noargs():
    with pytest.raises(ValueError):
        plot_slice()


def test_plot_slice_ct(sample_ct, tmp_path):
    plot_slice(ct=sample_ct, save_filename=str(tmp_path / "slice_ct.png"), show_plot=False)


def test_plot_slice_cst(sample_cst, tmp_path):
    plot_slice(cst=sample_cst, save_filename=str(tmp_path / "slice_cst.png"), show_plot=False)


def test_plot_slice_ct_cst(sample_ct, sample_cst, tmp_path):
    plot_slice(
        ct=sample_ct,
        cst=sample_cst,
        save_filename=str(tmp_path / "slice_ct_cst.png"),
        show_plot=False,
    )


def test_plot_slice_with_overlay(sample_ct, sample_cst, sample_overlay, tmp_path):
    plot_slice(
        ct=sample_ct,
        cst=sample_cst,
        overlay=sample_overlay,
        overlay_unit="Gy",
        view_slice=5,
        save_filename=str(tmp_path / "slice_with_overlay.png"),
        show_plot=False,
    )


def test_plot_slice_parameters_coronal(sample_ct, sample_cst, sample_overlay, tmp_path):
    plot_slice(
        ct=sample_ct,
        cst=sample_cst,
        overlay=sample_overlay,
        view_slice=[50, 2, 5],
        plane="coronal",
        overlay_unit="Gy",
        contour_line_width=2,
        overlay_alpha=0.2,
        overlay_rel_threshold=0.5,
        save_filename=str(tmp_path / "slice_parameters_coronal.png"),
        show_plot=False,
    )


def test_plot_slice_parameters_sagittal(sample_ct, sample_cst, sample_overlay, tmp_path):
    plot_slice(
        ct=sample_ct,
        cst=sample_cst,
        overlay=sample_overlay,
        plane="sagittal",
        save_filename=str(tmp_path / "parameters_sagittal.png"),
        show_plot=False,
    )


def test_plot_slice_parameters_invalid_plane(sample_ct, sample_cst, sample_overlay, tmp_path):
    with pytest.raises(ValueError):
        plot_slice(
            ct=sample_ct,
            cst=sample_cst,
            overlay=sample_overlay,
            plane="invalid_plane",
            show_plot=False,
        )


def test_plot_slice_multiple_slices_ct(sample_ct, tmp_path):
    plot_slice(
        ct=sample_ct,
        view_slice=[2, 5, 8],
        save_filename=str(tmp_path / "slice_multiple_slices_ct.png"),
        show_plot=False,
    )


def test_global_max(sample_ct, sample_cst, sample_overlay, tmp_path):
    sample_overlay = sitk.GetArrayFromImage(sample_overlay)
    sample_overlay[2] = sample_overlay[2] * 3
    plot_slice(
        ct=sample_ct,
        cst=sample_cst,
        overlay=sample_overlay,
        use_global_max=True,
        view_slice=[2, 5, 8],
        show_plot=False,
        save_filename=str(tmp_path / "global_max.png"),
    )


def test_plot_multiple_slices(sample_ct, sample_cst, sample_overlay, tmp_path):
    sample_overlay = sitk.GetArrayFromImage(sample_overlay)
    plot_slice(
        ct=sample_ct,
        cst=sample_cst,
        overlay=sample_overlay,
        view_slice=[2, 5, 8],
        show_plot=False,
        save_filename=str(tmp_path / "plot_multiple_slices.png"),
    )


def test_plot_multiple_slices_numpy(sample_ct, sample_cst, sample_overlay, tmp_path):
    sample_overlay = sitk.GetArrayFromImage(sample_overlay)
    plot_slice(
        ct=sample_ct,
        cst=sample_cst,
        overlay=sample_overlay,
        view_slice=np.array([2, 5, 8]),
        show_plot=False,
        save_filename=str(tmp_path / "plot_multiple_slices.png"),
    )



from abc import ABC
from typing import Any, Union
from typing_extensions import Annotated, Self
import warnings
from pydantic import (
    Field,
    field_validator,
    model_validator,
    computed_field,
    StringConstraints,
)

import numpy as np
import SimpleITK as sitk

from pyRadPlan.core import PyRadPlanBaseModel, np2sitk
from pyRadPlan.ct import CT

# Default overlap priorities
DEFAULT_OVERLAPS = {"TARGET": 0, "OAR": 5, "HELPER": 10, "EXTERNAL": 15}


class VOI(PyRadPlanBaseModel, ABC):
    """
    Represents a Volume of Interest (VOI).

    Parameters
    ----------
    name : str
        The name of the VOI.
    ct_image : CT
        The CT image where the VOI is defined.
    mask : np.ndarray or sitk.Image
        Boolean mask (using 0,1) for referencing of voxels
        (Multiple allocations possible for robust scenarios)
    alpha_x : float, optional
        The alpha_x value. Defaults to 0.1.
    beta_x : float, optional
        The beta_x value. Defaults to 0.05.
    overlap_priority : int
        The overlap priority of the VOI. Lowest number is overlapping higher numbers.
    """

    name: str
    ct_image: CT
    mask: sitk.Image
    alpha_x: float = Field(default=0.1)
    beta_x: float = Field(default=0.05)
    voi_type: Annotated[str, StringConstraints(strip_whitespace=True, to_upper=True)]

    overlap_priority: int = Field(
        alias="Priority", default_factory=lambda data: DEFAULT_OVERLAPS[data["voi_type"]]
    )

    # TODO: it would be nicer if this was a list of optimization.Objective, but that would create a
    # circular import. Forward type hinting does not work directly due to pydantic. If someone has
    # a better idea how to solve this, please do so.
    objectives: list[Any] = Field(default=[], description="List of objective function definitions")

    @field_validator("mask", mode="before")
    @classmethod
    def validate_mask_type(cls, v: Any) -> Any:
        """
        Validates the mask type.

        Parameters
        ----------
        v : Any
            The mask value to be validated.

        Returns
        -------
        sitk.Image
            The validated mask.

        Raises
        ------
        ValueError
            If the mask type is not supported.
        """
        if isinstance(v, np.ndarray):
            if v.dtype in ["bool", "int"]:
                v = v.astype("uint8")
            if v.dtype != "uint8":
                raise ValueError(
                    f"{v.dtype} is not supported for index mask. Please use uint8 or boolean mask."
                )

            if v.ndim == 3:
                return sitk.GetImageFromArray(v, False)

            if v.ndim == 4:
                mask = []
                for i in range(v.shape[0]):
                    mask.append(sitk.GetImageFromArray(v[i], False))
                v = sitk.JoinSeries(mask)
                return v

            raise ValueError("Dimensionality not supported!")

        if isinstance(v, sitk.Image):
            if sitk.GetArrayViewFromImage(v).dtype != "uint8":
                raise ValueError(
                    f"""{sitk.GetArrayViewFromImage(v).dtype} is not supported
                      for index mask. Please use uint8."""
                )
            return v

        raise ValueError("mask must be either passed as numpy array or SimpleITK image")

    @model_validator(mode="after")
    def validate_mask(self):
        """
        Check if the given indices are valid for the CT image.

        Raises
        ------
        ValueError
            If the mask is not a sitk.Image.
        ValueError
            If the dimensions of the mask do not match the CT image.
        """
        if not isinstance(self.mask, sitk.Image):
            raise ValueError("Sanity check failed - mask is not a SimpleITK image")

        # check dimensions of sitk image
        dims = self.mask.GetSize()
        if dims != self.ct_image.cube_hu.GetSize():
            raise ValueError(
                f"Mask provided with dimensions {dims}, "
                f"but ct has dimensions {self.ct_image.cube_hu.GetSize()}"
            )

        # set image parameters for mask in accordance to ct image
        self.mask.SetOrigin(self.ct_image.cube_hu.GetOrigin())
        self.mask.SetSpacing(self.ct_image.cube_hu.GetSpacing())
        self.mask.SetDirection(self.ct_image.cube_hu.GetDirection())
        return self

    @computed_field
    @property
    def indices(self) -> np.ndarray:
        """
        Returns the indices of the voxels in the mask using Fortran/SITK
        convention.

        Returns
        -------
        np.ndarray
            The indices of the voxels.
        """
        return np2sitk.sitk_mask_to_linear_indices(self.mask, order="sitk")

    @computed_field
    @property
    def indices_numpy(self) -> np.ndarray:
        """
        Returns the indices of the voxels in the mask using C/numpy
        convention.

        Returns
        -------
        np.ndarray
            The indices of the voxels.
        """
        return np2sitk.sitk_mask_to_linear_indices(self.mask, order="numpy")

    @computed_field
    @property
    def _numpy_mask(self) -> np.ndarray:
        """
        Returns the mask as a numpy array.

        Returns
        -------
        np.ndarray
            The mask as a numpy array.
        """
        return sitk.GetArrayViewFromImage(self.mask)

    @computed_field
    @property
    def num_of_scenarios(self) -> int:
        """
        Returns the number of scenarios.

        Returns
        -------
        int
            The number of scenarios.
        """

        if self.mask.GetDimension() == 4:
            return self.mask.GetSize()[3]

        return 1

    def get_indices(self, order="sitk") -> np.ndarray:
        """
        Returns the indices of the voxels in the mask.

        Parameters
        ----------
        order : str, optional
            The order of the indices. Defaults to "sitk".

        Returns
        -------
        np.ndarray
            The indices of the voxels.
        """
        if order == "numpy":
            return self.indices_numpy
        if order == "sitk":
            return self.indices
        raise ValueError(f"Unknown order: {order}")

    def scenario_indices(self, order_type="numpy") -> Union[np.ndarray, list[np.ndarray]]:
        """
        Returns the flattened indices of the individual scenarios.

        Parameters
        ----------
        order_type : str, optional
            The order type. Defaults to "numpy".

        Returns
        -------
        List[np.ndarray]
            The flattened indices of the individual scenarios.
        """
        if order_type == "numpy":
            _order = "C"
        elif order_type == "sitk":
            _order = "F"
        else:
            raise ValueError(f"Unknown order type: {order_type}")

        arr = sitk.GetArrayViewFromImage(self.mask)
        if len(arr.shape) == 3:
            return np.ravel_multi_index(np.argwhere(arr).T, dims=arr.shape, order=_order)
        if len(arr.shape) == 4:
            return [
                np.ravel_multi_index(np.argwhere(arr[i]).T, dims=arr[i].shape, order=_order)
                for i in range(arr.shape[0])
            ]

        raise ValueError("Sanity check failed - mask has invalid dimensions")

    def masked_ct(self, order_type="numpy") -> Union[sitk.Image, np.ndarray]:
        """
        Returns the masked CT image, either as a numpy array or a SimpleITK
        image.

        Parameters
        ----------
        order_type : str, optional
            The order type. Defaults to "numpy".

        Returns
        -------
        sitk.Image or np.ndarray
            The masked CT image.
        """

        if order_type not in ["numpy", "sitk"]:
            raise ValueError(f"Invalid order type requested: {order_type}")

        if len(self.mask.GetSize()) == 3:
            masked_ct = sitk.Mask(self.ct_image.cube_hu, self.mask)
        elif len(self.mask.GetSize()) == 4:
            masked_ct = [
                sitk.Mask(self.ct_image.cube_hu[:, :, :, i], self.mask[:, :, :, i])
                for i in range(self.mask.GetSize()[-1])
            ]
            masked_ct = sitk.JoinSeries(masked_ct)
        else:
            raise ValueError("Sanity check failed - mask has invalid dimensions")

        if order_type == "numpy":
            return sitk.GetArrayFromImage(masked_ct)
        if order_type == "sitk":
            return masked_ct

        raise ValueError(f"Sanity check failed -- Invalid order type requested: {order_type}")

    @computed_field
    @property
    def scenario_ct_data(self) -> Union[list[np.ndarray], np.ndarray]:
        """
        Returns a list of CT data for the individual scenarios.

        Returns
        -------
        List[np.ndarray]
            The CT data for the individual scenarios.
        """

        mask_np = sitk.GetArrayFromImage(self.mask).astype("bool")
        ct_np = sitk.GetArrayFromImage(self.ct_image.cube_hu)

        if len(self.mask.GetSize()) == 3:
            return ct_np[mask_np]

        if len(self.mask.GetSize()) == 4:
            return [ct_np[i][mask_np[i]] for i in range(mask_np.shape[0])]

        raise ValueError("Sanity Check failed -- Unsupported dimensionality of stored mask")

    def to_matrad(self, context: str = "mat-file") -> Any:
        """
        Creates an object that can be interpreted by matRad in the given
        context.

        Returns
        -------
        Any
            VOI as list to write cell arrays.
        """

        if context != "mat-file":
            raise ValueError(f"Context {context} not supported")

        voi_list = [0]  # We store an ID which will be changed by cst if exported from there
        voi_list.append(self.name)
        voi_list.append(self.voi_type)
        if self.num_of_scenarios == 1:
            index_lists = np.ndarray(shape=(1,), dtype=object)
            mask_array = sitk.GetArrayFromImage(self.mask)
            mask_array = np.swapaxes(mask_array, 1, 2)
            indices = np.argwhere(mask_array.ravel(order="C") > 0) + 1
            index_lists[0] = np.array(indices, dtype=float)

        else:
            index_lists = self.scenario_indices(order_type="numpy")
            for i, index_list in enumerate(index_lists):
                index_lists[i] = index_list.astype(float)

        voi_list.append(index_lists)

        property_dict = {
            "alphaX": self.alpha_x,
            "betaX": self.beta_x,
            "Priority": self.overlap_priority,
        }
        voi_list.append(property_dict)

        # Will not be populated in here but in cst if exported from there
        objective_dict = {}
        voi_list.append([objective_dict])

        return voi_list

    def resample_on_new_ct(self, new_ct: CT) -> Self:
        """
        Resample on new CT image.

        Parameters
        ----------
        new_ct : CT
            The new CT image to resample the VOI on.

        Returns
        -------
        Self
            The resampled VOI.
        """

        if not isinstance(new_ct, CT):
            raise ValueError("new_ct must be a CT object")

        if self.mask.GetDimension() == 3:
            new_mask = sitk.Resample(self.mask, new_ct.cube_hu)
        elif self.mask.GetDimension() == 4:
            new_mask = []
            for i in range(self.mask.GetSize()[-1]):
                new_mask.append(sitk.Resample(self.mask[:, :, :, i], new_ct.cube_hu))
            new_mask = sitk.JoinSeries(new_mask)
        else:
            raise ValueError("Sanity check failed -- mask has invalid dimensions")

        resampled_voi = self.model_copy(update={"mask": new_mask, "ct_image": new_ct})

        if len(resampled_voi.indices) == 0:
            warnings.warn("Resampling created an empty structure")

        return resampled_voi


class OAR(VOI):
    """
    Represents an organ at risk (OAR).

    Attributes
    ----------
    Inherits all attributes from Plan.

    Methods
    -------
    voi_type : str
        Returns the voi_type as 'OAR'.
    """

    voi_type: str = "OAR"

    @field_validator("voi_type", mode="after")
    @classmethod
    def validate_voi_type(cls, v: str) -> str:
        """
        Validates the voi type for an OAR.

        Parameters
        ----------
        v : str
            The voi type to be validated.

        Returns
        -------
        str
            The validated voi type.

        Raises
        ------
        ValueError
            If the voi type is not "OAR".
        """

        if v != "OAR":
            raise ValueError('VOI type for OAR must be "OAR"')
        return v


class Target(VOI):
    """
    Represents a target VOI.

    Attributes
    ----------
    Inherits all attributes from Plan.

    Methods
    -------
    voi_type : str
        Returns the voi_type as 'TARGET'.
    """

    voi_type: str = "TARGET"

    @field_validator("voi_type", mode="after")
    @classmethod
    def validate_voi_type(cls, v: str) -> str:
        """
        Validates the voi type for a Target.

        Parameters
        ----------
        v : str
            The voi type to be validated.

        Returns
        -------
        str
            The validated voi type.

        Raises
        ------
        ValueError
            If the voi type is not "OAR".
        """

        if v != "TARGET":
            raise ValueError('VOI type for a Target must be "TARGET"')
        return v


class HelperVOI(VOI):
    """
    Represents a helper VOI.

    Attributes
    ----------
    Inherits all attributes from Plan.

    Methods
    -------
    voi_type : str
        Returns the voi_type as 'HELPER'.
    """

    voi_type: str = "HELPER"

    @field_validator("voi_type", mode="after")
    @classmethod
    def validate_voi_type(cls, v: str) -> str:
        """
        Validates the voi type for a HelperVOI.

        Parameters
        ----------
        v : str
            The voi type to be validated.

        Returns
        -------
        str
            The validated voi type.

        Raises
        ------
        ValueError
            If the voi type is not "HELPER".
        """
        if v != "HELPER":
            raise ValueError('VOI type for a HelperVOI must be "HELPER"')
        return v


class ExternalVOI(VOI):
    """
    Represents an external contour limiting voxels to be considered for
    planning (EXTERNAL).

    Attributes
    ----------
    Inherits all attributes from Plan.

    Methods
    -------
    voi_type : str
        Returns the voi_type as 'EXTERNAL'.
    """

    voi_type: str = "EXTERNAL"

    @field_validator("voi_type", mode="after")
    @classmethod
    def validate_voi_type(cls, v: str) -> str:
        """
        Validates the voi type for an EXTERNAL contour.

        Parameters
        ----------
        v : str
            The voi type to be validated.

        Returns
        -------
        str
            The validated voi type.

        Raises
        ------
        ValueError
            If the voi type is not "EXTERNAL".
        """

        if v != "EXTERNAL":
            raise ValueError('VOI type for EXTERNAL must be "EXTERNAL"')
        return v


__VOITYPES__ = {"OAR": OAR, "TARGET": Target, "HELPER": HelperVOI, "EXTERNAL": ExternalVOI}


def create_voi(data: Union[dict[str, Any], VOI, None] = None, **kwargs) -> VOI:
    """
    Factory function to create a VOI object.

    Parameters
    ----------
    data : Union[dict[str, Any], VOI, None]
        Dictionary containing the data to create the VOI object.
    **kwargs
        Arbitrary keyword arguments.

    Returns
    -------
    VOI
        A VOI object.
    """

    if data:
        # If data is already a VOI object, return it directly
        if isinstance(data, VOI):
            return data

        # obtain voi type if we have a dict including camelCase check
        voi_type = data.get("voi_type", data.get("voiType", None))

        if voi_type in __VOITYPES__:
            return __VOITYPES__[voi_type].model_validate(data)

        raise ValueError(f"Invalid VOI type: {voi_type}")

    voi_type = kwargs.get("voi_type", "")

    if voi_type in __VOITYPES__:
        return __VOITYPES__[voi_type](**kwargs)

    raise ValueError(f"Invalid VOI type: {voi_type}")


def validate_voi(data: Union[dict[str, Any], VOI, None] = None, **kwargs) -> VOI:
    """
    Validates and creates a VOI object.
    Synonym to create_voi but should be used in validation context.

    Parameters
    ----------
    voi : Union[dict[str, Any], VOI, None], optional
        Dictionary containing the data to create the VOI object, by default None.
    **kwargs
        Arbitrary keyword arguments.

    Returns
    -------
    VOI
        A validated VOI object.
    """
    return create_voi(data, **kwargs)



"""Module defining various objective functions for optimization."""

from ._objective import Objective
from ._std import DoseUniformity
from ._eud import EUD
from ._max_dvh import MaxDVH
from ._mean import MeanDose
from ._min_dvh import MinDVH
from ._squared_dev import SquaredDeviation
from ._squared_over import SquaredOverdosing
from ._squared_under import SquaredUnderdosing

from ._factory import get_available_objectives, get_objective, register_objective

register_objective(SquaredDeviation)
register_objective(SquaredOverdosing)
register_objective(SquaredUnderdosing)
register_objective(MeanDose)
register_objective(EUD)
register_objective(MinDVH)
register_objective(MaxDVH)
register_objective(DoseUniformity)

__all__ = [
    "Objective",
    "DoseUniformity",
    "EUD",
    "MaxDVH",
    "MeanDose",
    "MinDVH",
    "SquaredDeviation",
    "SquaredOverdosing",
    "SquaredUnderdosing",
    "get_available_objectives",
    "get_objective",
    "register_objective",
]



import warnings
from typing import Any, Optional, Annotated, cast
from typing_extensions import Self
import numpy as np
from pydantic import (
    Field,
    StringConstraints,
    computed_field,
    field_validator,
    model_validator,
    ValidatorFunctionWrapHandler,
    ValidationInfo,
    ValidationError,
)
from numpydantic import NDArray, Shape
from pyRadPlan.core import PyRadPlanBaseModel
from ._base import ExternalBeamMachine


class PhotonSVDKernel(PyRadPlanBaseModel):
    """Kernel data for photon beams.

    Attributes
    ----------
    energy : float
        The maximum energy of the photon beam in MeV (corresponds to the LINAC voltage in MV).
    """

    # some basic kernel information
    energy: float = Field(ge=0.0, description="The energy of the photon beam in MeV.")
    m: float = Field(description="attuneation factor")
    penumbra: float = Field(
        default=5.0,
        description="The penumbra of an open field as FWHM at Isocenter",
        alias="penumbarFWHMatIso",
    )

    # Tabulated kernel data
    kernel_betas: NDArray[Shape["1-*"], np.float64] = Field(
        description="weighting factors for kernel components", alias="betas"
    )
    kernel_ssds: NDArray[Shape["1-*"], np.float64] = Field(description="Kernel SSD points")
    kernel_pos: NDArray[Shape["1-*"], np.float64] = Field(description="Kernel position")
    kernel_data: NDArray[Shape["1-*,1-*, 1-*"], np.float64] = Field(
        description="Spatial grid of kernels"
    )

    primary_fluence: NDArray[Shape["1-*,2"], np.float64] = Field(description="Primary fluence")

    @computed_field(return_type=int)
    @property
    def num_kernel_components(self):
        return self.kernel_data.shape[1]

    def get_kernels_at_ssd(self, ssd: float) -> NDArray[Shape["1-*, 1-*"], np.float64]:
        """Get the kernel components at a specific SSD value.

        Parameters
        ----------
        ssd : float
            The SSD value to search for.

        Returns
        -------
        NDArray[Shape["1-*, 1-*"], np.float64]
            The kernels at the specified SSD value.
        """

        # For performance reasons we slice kernel_data at the closest SSD value
        if ssd < self.kernel_ssds.min() or ssd > self.kernel_ssds.max():
            warnings.warn(f"SSD value {ssd} is not in the kernel SSD range.")

        ix = np.argmin(np.abs(self.kernel_ssds - ssd))
        return self.kernel_data[ix]

    def kernel_interpolator(self):
        pass

    @field_validator("kernel_betas", "kernel_ssds", "kernel_pos", "primary_fluence", mode="wrap")
    @classmethod
    def validate_kernel_data_shapes(
        cls, v: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
    ) -> NDArray:
        try:
            return handler(v, info)
        except ValidationError:
            try:
                v = np.asarray(v)
                return handler(v, info)
            except Exception as exc:
                raise ValueError("Input not an array") from exc

    @model_validator(mode="after")
    def validate_kernel_data(self) -> Self:
        if self.kernel_data.shape[1] != self.kernel_betas.shape[0]:
            raise ValueError("Kernel data and kernel betas do not match in length")

        if self.kernel_data.shape[0] != self.kernel_ssds.shape[0]:
            raise ValueError("Kernel data and kernel ssds do not match in length")

        if self.kernel_data.shape[2] != self.kernel_pos.shape[0]:
            raise ValueError("Kernel data and kernel positions do not match in length")

        return self


class PhotonLINAC(ExternalBeamMachine):
    """Base class for Machine objects
    Defines minimum meta-data a photon LINAC machine must hold
    Provides multiple data storage formats, currently supported
    - SVD Kernel Data (Bortfeld 1993, 10.1118/1.597070).

    Attributes
    ----------
    scd : float
        The source-to-collimator distance of the machine
    pb_kernels : dict[float, PhotonSVDKernel]
        Pencil-beam kernels for the machine
    of : np.ndarray
        Output factor data
    tpr : np.ndarray
        Tissue-to-phantom ratio data
    """

    radiation_mode: Annotated[str, StringConstraints(pattern="photons")] = Field(
        default="photons", validate_default=True
    )

    scd: float = Field(ge=0.0, description="Source to collimator distance", alias="SCD")

    # Optionally we can have pencil-beam kernels
    pb_kernels: Optional[dict[float, PhotonSVDKernel]] = None

    # Commissioning data
    of: Optional[NDArray] = Field(default=None, description="Output Factor")
    tpr: Optional[NDArray] = Field(default=None, description="Tissue-to-phantom-Ratio")

    # TODO: Beam model descriptions for Monte Carlo

    @classmethod
    def _parse_tabulated_energy_data_from_mat(
        cls, tabulated_energy_data: dict, returned_data: dict
    ):
        """Parse the tabulated energy data from a matRad machine file."""

        super(PhotonLINAC, cls)._parse_tabulated_energy_data_from_mat(
            tabulated_energy_data, returned_data
        )

        # parse kernel
        # returned_data["scd"] = tabulated_energy_data.get("scd", 0.0)
        returned_data["pb_kernels"] = tabulated_energy_data

    @field_validator("pb_kernels", mode="wrap")
    @classmethod
    def validate_pb_kernels(
        cls, v: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
    ) -> PhotonSVDKernel:
        try:
            return handler(v, info)
        except ValidationError as err:
            # If we have a dict, we try to convert it to a PhotonSVDKernel object
            if isinstance(v, dict):
                # This check tries to check if we have a matRad kenrel data set
                if "kernel" in v and "kernel_data" not in v:
                    try:
                        kernel = cast(dict, v["kernel"])
                        v["kernel_ssds"] = np.asarray(kernel.pop("SSD"), dtype=np.float64)
                        # matRad kernels are named consequtively kernel1, kernel2, ...
                        kernel_list = [
                            np.asarray(kernel.pop(f"kernel{i}"), dtype=np.float64)
                            for i in range(1, len(kernel) + 1)
                        ]
                        v["kernel_data"] = np.ascontiguousarray(
                            np.moveaxis(np.asarray(kernel_list), 0, 1)
                        )

                        # kernel dictionary by energy
                        v = {float(v["energy"]): v}
                    except Exception as exc:
                        raise ValueError("Could not parse matRad kernel data") from exc

                    return handler(v, info)

            # Otherwise we don't know what to do
            raise err

    def get_kernel_by_index(self, ix_energy: int) -> PhotonSVDKernel:
        """Get the pencil beam kernel for a specific energy index.

        Parameters
        ----------
        ix_energy : int
            Index of the energy level in the stored data

        Returns
        -------
        IonPencilBeamKernel
            Pencil beam kernel data
        """
        return self.get_kernel_by_energy(self.energies[ix_energy])

    def get_kernel_by_energy(self, energy: float) -> PhotonSVDKernel:
        """Get the pencil beam kernel for a specific energy value.

        Parameters
        ----------
        energy : float
            Energy value to search for

        Returns
        -------
        IonPencilBeamKernel
            Pencil beam kernel data
        """
        if self.pb_kernels is None:
            raise ValueError("No pencil beam kernels available for this machine.")
            # return None
        return self.pb_kernels[energy]



import pytest
from pyRadPlan.core import Grid
from pydantic import ValidationError
import numpy as np
import SimpleITK as sitk


def test_grid_invalid_resolution():
    with pytest.raises(ValueError):
        grid = Grid(resolution={"x": -1, "y": 1, "z": 1})


def test_grid_invalid_dimensions():
    with pytest.raises(ValueError):
        grid = Grid(dimensions=(0, 10, 10))


def test_grid_resolution_keys():
    grid = Grid(resolution={"x": 1, "y": 1, "z": 1}, dimensions=(10, 10, 10))
    assert grid.resolution.keys() == {"x", "y", "z"}


def test_grid_resolution_values():
    grid = Grid(resolution={"x": 1, "y": 1, "z": 1}, dimensions=(10, 10, 10))
    assert all(isinstance(value, float) for value in grid.resolution.values())


def test_grid_resolution_positive():
    grid = Grid(resolution={"x": 1, "y": 1, "z": 1}, dimensions=(10, 10, 10))
    assert all(value > 0 for value in grid.resolution.values())


def test_grid_resolution_vector():
    grid = Grid(resolution={"x": 1, "y": 1, "z": 1}, dimensions=(10, 10, 10))
    assert isinstance(grid.resolution_vector, np.ndarray)
    assert grid.resolution_vector.shape == (3,)
    assert all(isinstance(value, float) for value in grid.resolution_vector)
    assert np.isclose(
        grid.resolution_vector, np.array([v for v in grid.resolution.values()])
    ).all()


def test_grid_invalid_resolution_keys():
    with pytest.raises(ValidationError):
        grid = Grid(dimensions=(10, 10, 10), resolution={"a": 1, "b": 1, "c": 1})


def test_grid_invalid_resolution_negative():
    with pytest.raises(ValidationError):
        grid = Grid(dimensions=(10, 10, 10), resolution={"x": -1, "y": 1, "z": 1})


def test_grid_dimensions():
    grid = Grid(dimensions=(10, 10, 10), resolution={"x": 1, "y": 1, "z": 1})
    assert isinstance(grid.dimensions, tuple)
    assert len(grid.dimensions) == 3
    assert all(isinstance(value, int) for value in grid.dimensions)


def test_grid_num_voxels():
    grid = Grid(dimensions=(10, 10, 10), resolution={"x": 1, "y": 1, "z": 1})
    assert isinstance(grid.num_voxels, int)
    assert grid.num_voxels == 1000


def test_grid_to_matrad():
    grid = Grid(resolution={"x": 1, "y": 1, "z": 1}, dimensions=(10, 10, 10))
    matrad_dict = grid.to_matrad()
    assert isinstance(matrad_dict, dict)
    assert "dimensions" in matrad_dict
    assert "numOfVoxels" in matrad_dict
    assert isinstance(matrad_dict["dimensions"], tuple)
    assert isinstance(matrad_dict["numOfVoxels"], float)
    assert matrad_dict["numOfVoxels"] == 1000.0


def test_grid_from_sitk_image():
    # Create a SimpleITK image
    image = sitk.Image(10, 10, 10, sitk.sitkFloat32)
    image.SetSpacing((1.0, 1.0, 1.0))
    image.SetOrigin((0.0, 0.0, 0.0))
    image.SetDirection(np.eye(3).flatten())

    # Create a Grid object from the SimpleITK image
    grid = Grid.from_sitk_image(image)

    # Check the Grid object properties
    assert grid.resolution == {"x": 1.0, "y": 1.0, "z": 1.0}
    assert grid.dimensions == (10, 10, 10)
    assert all(grid.origin == (0.0, 0.0, 0.0))
    assert np.array_equal(grid.direction, np.eye(3))


@pytest.fixture
def sample_grid():
    """Create a sample Grid object for testing."""
    resolution = {"x": 1.0, "y": 1.0, "z": 2.0}
    dimensions = (50, 100, 100)
    origin = np.array([0.0, 0.0, 0.0], dtype=np.float64)
    direction = np.eye(3, dtype=np.float64)
    return Grid(resolution=resolution, dimensions=dimensions, origin=origin, direction=direction)


def test_resample_grid_with_dict(sample_grid):
    target_resolution = {"x": 2.0, "y": 2.0, "z": 2.0}
    resampled_grid = sample_grid.resample(target_resolution)
    assert resampled_grid.resolution == target_resolution
    assert resampled_grid.dimensions == (25, 50, 100)
    assert np.allclose(resampled_grid.origin, [0.0, 0.0, 0.0])


def test_resample_grid_with_ndarray(sample_grid):
    target_resolution = np.array([2.0, 2.0, 2.0])
    resampled_grid = sample_grid.resample(target_resolution)
    assert resampled_grid.resolution == {"x": 2.0, "y": 2.0, "z": 2.0}
    assert resampled_grid.dimensions == (25, 50, 100)
    assert np.allclose(resampled_grid.origin, [0.0, 0.0, 0.0])


def test_resample_grid_irregular_resolution(sample_grid):
    target_resolution = {"x": 2.0, "y": 5.0, "z": 3.0}
    resampled_grid = sample_grid.resample(target_resolution)
    assert resampled_grid.resolution == target_resolution
    assert resampled_grid.dimensions == (25, 20, 67)
    assert np.allclose(resampled_grid.origin, [0.0, 0.0, -0.5])


def test_resample_grid_with_shifted_rotated_grid(sample_grid):
    sample_grid.origin = np.array([10.0, 10.0, 10.0], dtype=np.float64)
    angle = np.pi / 4  # 45 degrees
    cos_angle = np.cos(angle)
    sin_angle = np.sin(angle)
    sample_grid.direction = np.array(
        [[cos_angle, -sin_angle, 0], [sin_angle, cos_angle, 0], [0, 0, 1]], dtype=np.float64
    )
    target_resolution = {"x": 2.0, "y": 2.0, "z": 2.0}
    resampled_grid = sample_grid.resample(target_resolution)
    assert resampled_grid.resolution == target_resolution
    assert resampled_grid.dimensions == (25, 50, 100)
    assert np.allclose(resampled_grid.origin, np.array([10.0, 10.0, 10.0]))


def test_resample_grid_invalid_resolution(sample_grid):
    with pytest.raises(ValueError):
        sample_grid.resample([2.0, 2.0])  # Invalid shape

    with pytest.raises(ValueError):
        sample_grid.resample("invalid")  # Invalid type


# test the custom implementation of the eq/neq operator on grids
def test_grids_equal():
    grid = Grid(resolution={"x": 1, "y": 1, "z": 1}, dimensions=(10, 10, 10))
    grid2 = Grid(resolution={"x": 1, "y": 1, "z": 1}, dimensions=(10, 10, 10))

    assert grid == grid2


def test_grids_not_equal():
    grid = Grid(resolution={"x": 1, "y": 1, "z": 1}, dimensions=(10, 10, 10))
    grid2 = Grid(resolution={"x": 1, "y": 1, "z": 2}, dimensions=(10, 10, 10))
    grid3 = Grid(resolution={"x": 1, "y": 1, "z": 1}, dimensions=(10, 10, 20))

    assert grid != grid2
    assert grid != grid3


def test_grids_not_equal_types():
    grid = Grid(resolution={"x": 1, "y": 1, "z": 1}, dimensions=(10, 10, 10))
    assert grid != 1
    assert grid != "string"
    assert grid != [1, 2, 3]

    # revert
    assert 1 != grid
    assert "string" != grid
    assert [1, 2, 3] != grid



import pytest
from pyRadPlan.scenarios import NominalScenario
from pyRadPlan.plan import create_pln, PhotonPlan, IonPlan


def test_create_pln_no_args():
    with pytest.raises(ValueError):
        create_pln()


def test_ionPlnEmptyConstructor():
    plan = IonPlan()
    assert plan.radiation_mode == "protons"


def test_photonPlnEmptyConstructor():
    plan = PhotonPlan()
    assert plan.radiation_mode == "photons"


def test_create_pln_dict_photons():
    plan = create_pln({"radiation_mode": "photons"})
    assert isinstance(plan, PhotonPlan)
    assert plan.radiation_mode == "photons"


def test_create_pln_from_Plan():
    plan = PhotonPlan()
    new_plan = create_pln(plan)
    assert isinstance(new_plan, PhotonPlan)

    plan = IonPlan()
    new_plan = create_pln(plan)
    assert isinstance(new_plan, IonPlan)


def test_create_pln_dict_ions():
    plan = create_pln({"radiation_mode": "protons"})
    assert isinstance(plan, IonPlan)
    assert plan.radiation_mode == "protons"

    plan = create_pln({"radiation_mode": "carbon"})
    assert isinstance(plan, IonPlan)
    assert plan.radiation_mode == "carbon"


def test_create_pln_dict_unknown():
    with pytest.raises(ValueError):
        create_pln({"radiation_mode": "unknown"})


def test_create_pln_kwargs_photon():
    plan = create_pln(radiation_mode="photons")
    assert isinstance(plan, PhotonPlan)
    assert plan.radiation_mode == "photons"


def test_create_pln_kwargs_ions():
    plan = create_pln(radiation_mode="protons")
    assert isinstance(plan, IonPlan)
    assert plan.radiation_mode == "protons"

    plan = create_pln(radiation_mode="carbon")
    assert isinstance(plan, IonPlan)
    assert plan.radiation_mode == "carbon"


def test_create_pln_kwargs_unknown():
    with pytest.raises(ValueError):
        create_pln(radiation_mode="unknown")


def test_create_pln_dict_photons_snake():
    scen_dict = NominalScenario().model_dump()

    pln_dict = {
        "radiation_mode": "photons",  # either photons / protons / carbon
        "machine": "Generic",
        "num_of_fractions": 30,
        "prescribed_dose": 60.0,
        "prop_stf": {},
        # dose calculation settings
        "prop_dose_calc": {},
        # optimization settings
        "prop_opt": {},
        "prop_seq": {},
        "mult_scen": scen_dict,
    }

    pln = create_pln(pln_dict)
    assert isinstance(pln, PhotonPlan)
    assert pln.radiation_mode == "photons"
    assert pln.num_of_fractions == 30
    assert pln.machine == "Generic"
    assert pln.prescribed_dose == 60.0
    assert isinstance(pln.mult_scen, NominalScenario)

    pln_from_dict = pln.model_dump()
    # print(set(pln_dict) ^ set(pln_from_dict))
    pln_dict.pop("mult_scen")
    pln_from_dict.pop("mult_scen")
    assert pln_dict == pln_from_dict


def test_create_pln_dict_photons_camel():
    scen = NominalScenario()
    scen_dict_camel = scen.to_matrad()
    scen_dict_snake = scen.model_dump()

    pln_dict_camel = {
        "radiationMode": "photons",  # either photons / protons / carbon
        "machine": "Generic",
        "numOfFractions": 30,
        "prescribedDose": 60.0,
        "propStf": {},
        # dose calculation settings
        "propDoseCalc": {},
        # optimization settings
        "propOpt": {},
        "propSeq": {},
        "multScen": scen_dict_camel,
    }

    pln_dict_snake = {
        "radiation_mode": "photons",  # either photons / protons / carbon
        "machine": "Generic",
        "num_of_fractions": 30,
        "prescribed_dose": 60.0,
        "prop_stf": {},
        # dose calculation settings
        "prop_dose_calc": {},
        # optimization settings
        "prop_opt": {},
        "prop_seq": {},
        "mult_scen": scen_dict_snake,
    }

    pln = create_pln(pln_dict_camel)
    assert isinstance(pln, PhotonPlan)
    assert pln.radiation_mode == "photons"
    assert pln.num_of_fractions == 30
    assert pln.machine == "Generic"
    assert pln.prescribed_dose == 60.0
    assert isinstance(pln.mult_scen, NominalScenario)

    pln_dict_snake.pop("mult_scen")
    pln_dict_camel.pop("multScen")

    pln_to_dict = pln.model_dump()
    pln_to_dict.pop("mult_scen")
    assert pln_dict_snake == pln_to_dict

    pln_to_dict_camel = pln.to_matrad()
    pln_to_dict_camel.pop("multScen")
    print(set(pln_dict_camel) ^ set(pln_to_dict_camel))
    assert pln_dict_camel == pln_to_dict_camel


def test_plan_to_matrad():
    plan = PhotonPlan(num_of_fractions=30)
    pln = plan.to_matrad()

    assert "numOfFractions" in pln
    # assert isinstance(pln["numOfFractions"],float)
    assert pln["numOfFractions"] == float(30)



import pytest

import numpy as np
from scipy.sparse import csc_array

from pyRadPlan.dij import Dij
from pyRadPlan.quantities import LETxDose, RTQuantity


@pytest.fixture
def sample_base_dij_dict():
    dij_dict = {
        "ct_grid": {
            "resolution": {"x": 1.5, "y": 1.5, "z": 1.5},
            "dimensions": (10, 10, 10),
            "num_of_voxels": 1000,
        },
        "dose_grid": {
            "resolution": {"x": 3.0, "y": 3.0, "z": 3.0},
            "dimensions": (5, 5, 5),
            "num_of_voxels": 125,
        },
        "num_of_beams": 1,
        "total_num_of_bixels": 10,
        "let_dose": np.empty((1, 1, 1), dtype=object),
        "bixel_num": np.arange(10),
        "ray_num": np.arange(10),
        "beam_num": np.zeros((10,), dtype=np.int64),
    }
    return dij_dict


@pytest.fixture
def sample_dij_dense(sample_base_dij_dict):
    sample_base_dij_dict["let_dose"].flat[0] = np.ones((125, 10), dtype=np.float32)
    dij = Dij.model_validate(sample_base_dij_dict)
    return dij


@pytest.fixture
def sample_dij_sparse(sample_base_dij_dict):
    dense_mat = np.ones((125, 10), dtype=np.float32)
    dense_mat[:100] = 0
    np.random.shuffle(dense_mat)
    sample_base_dij_dict["let_dose"].flat[0] = csc_array(dense_mat)
    dij = Dij.model_validate(sample_base_dij_dict)
    return dij


def test_LETxDose_constructor(sample_dij_dense):
    let_dose = LETxDose(sample_dij_dense)
    assert isinstance(let_dose, RTQuantity)
    assert let_dose.scenarios == [0]
    assert let_dose._dij == sample_dij_dense
    assert let_dose.dim == 1
    assert format(let_dose.unit, "~") == "Gy * µm / keV"
    assert let_dose.identifier == "let_dose"
    assert let_dose.name == "LETxDose"


def test_let_dose_dense(sample_dij_dense):
    let_dose = LETxDose(sample_dij_dense)

    fluence = range(10)
    ret_callable = let_dose(fluence)
    assert np.array_equal(let_dose._w_cache, fluence)
    ret_compute = let_dose.compute(fluence)

    assert isinstance(ret_callable, np.ndarray)
    assert ret_callable.dtype == sample_dij_dense.let_dose.dtype
    assert ret_callable.shape == sample_dij_dense.let_dose.shape

    dij_mat = sample_dij_dense.let_dose.flat[0]
    assert np.allclose(ret_callable.flat[0], dij_mat @ fluence)
    assert np.array_equal(ret_callable.flat[0], ret_compute.flat[0])

    ret_deriv = let_dose.compute_chain_derivative(np.ones(125, dtype=np.float32), fluence)
    assert np.array_equal(let_dose._w_grad_cache, fluence)
    assert np.array_equal(let_dose._qgrad_cache.flat[0], ret_deriv.flat[0])
    assert isinstance(ret_deriv, np.ndarray)
    assert ret_deriv.dtype == sample_dij_dense.let_dose.dtype
    assert ret_deriv.shape == sample_dij_dense.let_dose.shape
    assert np.allclose(ret_deriv.flat[0], dij_mat.T @ np.ones(125, dtype=np.float32))


def test_let_dose_sparse(sample_dij_sparse):
    let_dose = LETxDose(sample_dij_sparse)

    fluence = range(10)
    ret_callable = let_dose(fluence)
    assert np.array_equal(let_dose._w_cache, fluence)
    ret_compute = let_dose.compute(fluence)

    assert isinstance(ret_callable, np.ndarray)
    assert ret_callable.dtype == sample_dij_sparse.let_dose.dtype
    assert ret_callable.shape == sample_dij_sparse.let_dose.shape

    dij_mat = sample_dij_sparse.let_dose.flat[0]
    assert np.allclose(ret_callable.flat[0], dij_mat @ fluence)
    assert np.array_equal(ret_callable.flat[0], ret_compute.flat[0])

    ret_deriv = let_dose.compute_chain_derivative(np.ones(125, dtype=np.float32), fluence)
    assert np.array_equal(let_dose._w_grad_cache, fluence)
    assert np.array_equal(let_dose._qgrad_cache.flat[0], ret_deriv.flat[0])
    assert isinstance(ret_deriv, np.ndarray)
    assert ret_deriv.dtype == sample_dij_sparse.let_dose.dtype
    assert ret_deriv.shape == sample_dij_sparse.let_dose.shape
    assert np.allclose(ret_deriv.flat[0], dij_mat.T @ np.ones(125, dtype=np.float32))



from ._engine import MatRadEngine


class MatRadEngineMatlab(MatRadEngine):
    @property
    def engine(self):
        return self.__engine

    @property
    def matRadVersionStr(self):
        return self.__matRadVersionStr

    @property
    def engineVersionStr(self):
        return self.__engineVersionStr

    def __init__(self, matRadPath="./matRad"):
        import matlab.engine

        self.matRadPath = matRadPath
        self.__engine = matlab.engine.start_matlab()
        self.__engineVersionStr = self.engine.version()
        self.__engine.addpath(self.engine.genpath(self.matRadPath), nargout=0)
        self.__engine.matRad_rc
        self.__matRadVersionStr = self.engine.matRad_version()
        # print(self.matRadVersionStr)
